from snovault.schema_utils import VALIDATOR_REGISTRY
from snovault import (
    COLLECTIONS,
    CONNECTION,
    Collection,
    ROOT,
    Root,
    calculated_property,
    collection,
    load_schema,
)
from snovault.resource_views import item_view_page
from .base import (
    ALLOW_EVERYONE_VIEW,
    SharedItem,
    ONLY_ADMIN_VIEW,
)
from pyramid.location import lineage
from pyramid.threadlocal import get_current_request
from pyramid.traversal import (
    find_resource,
)
from pyramid.view import view_config


@collection(
    name='pages',
    unique_key='page:location',
    properties={
        'title': 'Pages',
        'description': 'Portal pages',
    })
class Page(SharedItem):
    item_type = 'page'
    schema = load_schema('snowflakes:schemas/page.json')
    name_key = 'name'

    embedded = [
        'layout.blocks.image',
    ]

    def unique_keys(self, properties):
        keys = super(Page, self).unique_keys(properties)
        parent = properties.get('parent')
        name = properties['name']
        value = name if parent is None else u'{}:{}'.format(parent, name)
        keys.setdefault('page:location', []).append(value)
        return keys

    @calculated_property(
        condition=lambda context, request: request.resource_path(context.__parent__) == '/pages/',
        schema={
            "title": "Canonical URI",
            "type": "string",
        })
    def canonical_uri(self, name):
        if name == 'homepage':
            return '/'
        return '/%s/' % name

    @property
    def __parent__(self):
        parent_uuid = self.properties.get('parent')
        name = self.__name__
        collections = self.registry[COLLECTIONS]
        connection = self.registry[CONNECTION]
        if parent_uuid:  # explicit parent
            return connection.get_by_uuid(parent_uuid)
        elif name in collections or name == 'homepage':
            # collection default page; use pages collection as canonical parent
            return self.collection
        else:  # top level
            return self.registry[ROOT]

    def is_default_page(self):
        name = self.__name__
        collections = self.registry[COLLECTIONS]
        if self.properties.get('parent'):
            return False
        return name in collections or name == 'homepage'

    # Handle traversal to nested pages

    def __getitem__(self, name):
        resource = self.get(name)
        if resource is None:
            raise KeyError(name)
        return resource

    def __contains__(self, name):
        return self.get(name, None) is not None

    def get(self, name, default=None):
        location = str(self.uuid) + ':' + name
        connection = self.registry[CONNECTION]
        resource = connection.get_by_unique_key('page:location', location)
        if resource is not None:
            return resource
        return default

    def __resource_url__(self, request, info):
        # Record ancestor uuids in linked_uuids so renames of ancestors
        # invalidate linking objects.
        for obj in lineage(self):
            uuid = getattr(obj, 'uuid', None)
            if uuid is not None:
                request._linked_uuids.add(str(uuid))
        return None


def isNotCollectionDefaultPage(value, schema):
    if value:
        request = get_current_request()
        page = find_resource(request.root, value)
        if page.is_default_page():
            return 'You may not place pages inside an object collection.'

VALIDATOR_REGISTRY['isNotCollectionDefaultPage'] = isNotCollectionDefaultPage


@view_config(context=Page, permission='view', request_method='GET', name='page')
def page_view_page(context, request):
    # Embedding of items has to happen here as we don't know which of their subobjects
    # need embedding as we don't know the type and may need their full page view.
    properties = item_view_page(context, request)
    blocks = properties.get('layout', {}).get('blocks', [])
    for block in blocks:
        if 'item' in block and block['item']:
            block['item'] = request.embed(block['item'], '@@page', as_user=True)
    return properties


@calculated_property(
    category='page',
    name='default_page',
    context=Collection,
    schema={
        "title": "Default page",
        "type": "string",
        "linkTo": "Page",
    })
def collection_default_page(context, request):
    try:
        return request.embed('/pages/%s/@@page' % context.__name__, as_user=True)
    except KeyError:
        pass


@calculated_property(
    category='page',
    name='default_page',
    context=Root,
    schema={
        "title": "Default page",
        "type": "string",
        "linkTo": "Page",
    })
def root_default_page(context, request):
    try:
        return request.embed('/pages/homepage/@@page', as_user=True)
    except KeyError:
        pass



"""\
Example.

To load the initial data:

    %(prog)s production.ini

"""
from pyramid.paster import get_app
from functools import reduce
from snovault import (
    COLLECTIONS,
    TYPES,
)
from snovault.schema_utils import combine_schemas
from .interfaces import (
    ELASTIC_SEARCH,
    RESOURCES_INDEX,
)
import collections
import json
import logging


log = logging.getLogger(__name__)


EPILOG = __doc__

log = logging.getLogger(__name__)

# An index to store non-content metadata
META_MAPPING = {
    '_all': {
        'enabled': False,
        'analyzer': 'snovault_index_analyzer',
        'search_analyzer': 'snovault_search_analyzer'
    },
    'dynamic_templates': [
        {
            'store_generic': {
                'match': '*',
                'mapping': {
                    'index': False,
                    'store': True,
                },
            },
        },
    ],
}


PATH_FIELDS = ['submitted_file_name']
KEYWORD_FIELDS = ['schema_version', 'uuid', 'accession', 'alternate_accessions',
                  'aliases', 'status', 'date_created', 'submitted_by',
                  'internal_status', 'target', 'biosample_type']
TEXT_FIELDS = ['pipeline_error_detail', 'description', 'notes']


def sorted_pairs_hook(pairs):
    return collections.OrderedDict(sorted(pairs))


def sorted_dict(d):
    return json.loads(json.dumps(d), object_pairs_hook=sorted_pairs_hook)


def schema_mapping(name, schema):
    # If a mapping is explicitly defined, use it
    if 'mapping' in schema:
        return schema['mapping']

    if 'linkFrom' in schema:
        type_ = 'string'
    else:
        type_ = schema['type']

    # Elasticsearch handles multiple values for a field
    if type_ == 'array':
        return schema_mapping(name, schema['items'])

    if type_ == 'object':
        properties = {}
        for k, v in schema.get('properties', {}).items():
            mapping = schema_mapping(k, v)
            if mapping is not None:
                properties[k] = mapping
        return {
            'type': 'object',
            'include_in_all': False,
            'properties': properties,
        }

    if type_ == ["number", "string"]:
        return {
            'type': 'keyword',
            'copy_to': [],
            'fields': {
                'value': {
                    'type': 'float',
                    'ignore_malformed': True,
                },
                'raw': {
                    'type': 'keyword',
                }
            }
        }

    if type_ == 'boolean':
        return {
            'type': 'boolean',
            'store': True,
            'fields': {
                'raw': {
                    'type': 'keyword',
                }
            }
        }

    if type_ == 'string':
        if name in KEYWORD_FIELDS:
            field_type = 'keyword'
        elif name in TEXT_FIELDS:
            field_type = 'text'
        else:
            field_type = 'keyword'
        sub_mapping = {
            'type': field_type
        }
        return sub_mapping

    if type_ == 'number':
        return {
            'type': 'float',
            'store': True,
            'fields': {
                'raw': {
                    'type': 'keyword',
                }
            }
        }

    if type_ == 'integer':
        return {
            'type': 'long',
            'store': True,
            'fields': {
                'raw': {
                    'type': 'keyword',
                }
            }
        }


def index_settings():
    return {
        'settings': {
            'index.max_result_window': 99999,
            'index.mapping.total_fields.limit': 5000,
            'index.number_of_shards': 5,
            'index.number_of_replicas': 2,
            'analysis': {
                'filter': {
                    'substring': {
                        'type': 'edge_ngram',
                        'min_gram': 1,
                        'max_gram': 33
                    },
                    'english_stop': {
                        'type': 'stop',
                        'stopwords': '_english_'
                    },
                    'english_stemmer': {
                        'type': 'stemmer',
                        'language': 'english'
                    },
                    'english_possessive_stemmer': {
                        'type': 'stemmer',
                        'language': 'possessive_english'
                    },
                    'delimiter': {
                        'type': 'word_delimiter',
                        'catenate_all': True,
                        'preserve_original': True,
                        'stem_english_possessive': True,
                        'split_on_numerics': False
                    }
                },
                'analyzer': {
                    'default': {
                        'type': 'custom',
                        'tokenizer': 'whitespace',
                        'char_filter': 'html_strip',
                        'filter': [
                            'english_possessive_stemmer',
                            'lowercase',
                            'english_stop',
                            'english_stemmer',
                            'asciifolding',
                            'delimiter'
                        ]
                    },
                    'snovault_index_analyzer': {
                        'type': 'custom',
                        'tokenizer': 'whitespace',
                        'char_filter': 'html_strip',
                        'filter': [
                            'english_possessive_stemmer',
                            'lowercase',
                            'english_stop',
                            'english_stemmer',
                            'asciifolding',
                            'delimiter',
                            'substring'
                        ]
                    },
                    'snovault_search_analyzer': {
                        'type': 'custom',
                        'tokenizer': 'whitespace',
                        'filter': [
                            'english_possessive_stemmer',
                            'lowercase',
                            'english_stop',
                            'english_stemmer',
                            'asciifolding',
                            'delimiter'
                        ]
                    },
                    'snovault_path_analyzer': {
                        'type': 'custom',
                        'tokenizer': 'snovault_path_tokenizer',
                        'filter': ['lowercase']
                    }
                },
                'tokenizer': {
                    'snovault_path_tokenizer': {
                        'type': 'path_hierarchy',
                        'reverse': True
                    }
                }
            }
        }
    }


def audit_mapping():
    return {
        'category': {
            'type': 'keyword',
        },
        'detail': {
            'type': 'text',
            'index': 'true',
        },
        'level_name': {
            'type': 'keyword',
        },
        'level': {
            'type': 'integer',
        }
    }


def es_mapping(mapping):
    return {
        '_all': {
            'enabled': True,
            'analyzer': 'snovault_index_analyzer',
            'search_analyzer': 'snovault_search_analyzer'
        },
        'dynamic_templates': [
            {
                'template_principals_allowed': {
                    'path_match': "principals_allowed.*",
                    'match_mapping_type': "string",
                    'mapping': {
                        'type': 'keyword',
                        'include_in_all': False,
                    },
                },
            },
            {
                'template_unique_keys': {
                    'path_match': "unique_keys.*",
                    'match_mapping_type': "string",
                    'mapping': {
                        'type': 'keyword',
                    },
                },
            },
            {
                'template_links': {
                    'path_match': "links.*",
                    'match_mapping_type': "string",
                    'mapping': {
                        'type': 'keyword',
                        'include_in_all': False,
                    },
                },
            },
            {
                'strings': {
                    'match_mapping_type': "string",
                    'mapping': {
                        'type': 'keyword',
                        'include_in_all': False,
                    },
                },
            },
            {
                'integers': {
                    'match_mapping_type': "long",
                    'mapping': {
                        'type': 'long',
                        'fields': {
                            'raw': {
                                'type': 'keyword'
                            }
                        },
                       'include_in_all': False,
                    },
                },
            }
        ],
        'properties': {
            'uuid': {
                'type': 'keyword',
                'include_in_all': False,
            },
            'tid': {
                'type': 'keyword',
                'include_in_all': False,
            },
            'item_type': {
                'type': 'keyword',
            },
            'embedded': mapping,
            'object': {
                'type': 'object',
                'enabled': False,
                'include_in_all': False,
            },
            'properties': {
                'type': 'object',
                'enabled': False,
                'include_in_all': False,
            },
            'propsheets': {
                'type': 'object',
                'enabled': False,
                'include_in_all': False,
            },
            'embedded_uuids': {
                'type': 'keyword',
                'include_in_all': False,
            },
            'linked_uuids': {
                'type': 'keyword',
                'include_in_all': False,
            },
            'paths': {
                'type': 'keyword',
                'include_in_all': False,
            },
            'audit': {
                'type': 'object',
                'include_in_all': False,
                'properties': {
                    'ERROR': {
                        'type': 'object',
                        'properties': audit_mapping()
                    },
                    'NOT_COMPLIANT': {
                        'type': 'object',
                        'properties': audit_mapping()
                    },
                    'WARNING': {
                        'type': 'object',
                        'properties': audit_mapping()
                    },
                    'INTERNAL_ACTION': {
                        'type': 'object',
                        'properties': audit_mapping()
                    },
                },
            }
        }
    }


def combined_mapping(types, *item_types):
    combined = {
        'type': 'object',
        'properties': {},
    }
    for item_type in item_types:
        schema = types[item_type].schema
        mapping = schema_mapping(item_type, schema)
        for k, v in mapping['properties'].items():
            if k in combined:
                assert v == combined[k]
            else:
                combined[k] = v

    return combined


def type_mapping(types, item_type, embed=True):
    type_info = types[item_type]
    schema = type_info.schema
    mapping = schema_mapping(item_type, schema)
    if not embed:
        return mapping
    for prop in type_info.embedded_paths:
        s = schema
        m = mapping

        for p in prop.split('.'):
            ref_types = None

            subschema = s.get('properties', {}).get(p)
            if subschema is None:
                msg = 'Unable to find schema for %r embedding %r in %r' % (p, prop, item_type)
                raise ValueError(msg)

            subschema = subschema.get('items', subschema)
            if 'linkFrom' in subschema:
                _ref_type, _ = subschema['linkFrom'].split('.', 1)
                ref_types = [_ref_type]
            elif 'linkTo' in subschema:
                ref_types = subschema['linkTo']
                if not isinstance(ref_types, list):
                    ref_types = [ref_types]

            if ref_types is None:
                m = m['properties'][p]
                s = subschema
                continue

            s = reduce(combine_schemas, (types[t].schema for t in ref_types))

            # Check if mapping for property is already an object
            # multiple subobjects may be embedded, so be carful here
            if m['properties'][p]['type'] in ['keyword', 'text']:
                m['properties'][p] = schema_mapping(p, s)

            m = m['properties'][p]

    boost_values = schema.get('boost_values', None)
    if boost_values is None:
        boost_values = {
            prop_name: 1.0
            for prop_name in ['@id', 'title']
            if prop_name in mapping['properties']
        }
    for name, boost in boost_values.items():
        props = name.split('.')
        last = props.pop()
        new_mapping = mapping['properties']
        for prop in props:
            new_mapping = new_mapping[prop]['properties']
        new_mapping[last]['boost'] = boost
        new_mapping[last]['include_in_all'] = True
    return mapping


def create_elasticsearch_index(es, index, body):
    es.indices.create(index=index, body=body, wait_for_active_shards=1, ignore=[400, 404], master_timeout='5m', request_timeout=300)


def set_index_mapping(es, index, doc_type, mapping):
    es.indices.put_mapping(index=index, doc_type=doc_type, body=mapping, ignore=[400], request_timeout=300)


def create_snovault_index_alias(es, indices):
    es.indices.put_alias(index=','.join(indices), name=RESOURCES_INDEX, request_timeout=300)


def run(app, collections=None, dry_run=False):
    index = app.registry.settings['snovault.elasticsearch.index']
    registry = app.registry
    if not dry_run:
        es = app.registry[ELASTIC_SEARCH]
        print('CREATE MAPPING RUNNING')

    if not collections:
        collections = ['meta'] + list(registry[COLLECTIONS].by_item_type.keys())

    indices = []
    for collection_name in collections:
        if collection_name == 'meta':
            doc_type = 'meta'
            mapping = META_MAPPING
        else:
            index = doc_type = collection_name
            collection = registry[COLLECTIONS].by_item_type[collection_name]
            mapping = es_mapping(type_mapping(registry[TYPES], collection.type_info.item_type))

        if mapping is None:
            continue  # Testing collections
        if dry_run:
            print(json.dumps(sorted_dict({index: {doc_type: mapping}}), indent=4))
            continue
        create_elasticsearch_index(es, index, index_settings())
        set_index_mapping(es, index, doc_type, {doc_type: mapping})
        if collection_name != 'meta':
            indices.append(index)

    create_snovault_index_alias(es, indices)


def main():
    import argparse
    parser = argparse.ArgumentParser(
        description="Create Elasticsearch mapping", epilog=EPILOG,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument('--item-type', action='append', help="Item type")
    parser.add_argument('--app-name', help="Pyramid app name in configfile")
    parser.add_argument(
        '--dry-run', action='store_true', help="Don't post to ES, just print")
    parser.add_argument('config_uri', help="path to configfile")
    args = parser.parse_args()

    logging.basicConfig()
    app = get_app(args.config_uri, args.app_name)

    # Loading app will have configured from config file. Reconfigure here:
    logging.getLogger('snovault').setLevel(logging.DEBUG)

    return run(app, args.item_type, args.dry_run)


if __name__ == '__main__':
    main()



from past.builtins import basestring
from .typedsheets import cast_row_values
from functools import reduce
import io
import logging
import os.path

text = type(u'')

logger = logging.getLogger('snovault')
logger.setLevel(logging.INFO)  # doesn't work to shut off sqla INFO

ORDER = [
    'user',
    'award',
    'lab',
    'snowball',
    'snowfort',
    'snowflake',
    'image',
    'page'
]

IS_ATTACHMENT = [
    'attachment',
]

##############################################################################
# Pipeline components
#
# http://www.stylight.com/Numbers/pipes-and-filters-architectures-with-python-generators/
#
# Dictionaries are passed through the pipeline. By convention, values starting
# with an underscore (_) are ignored by the component posting a final value
# so are free for communicating information down the pipeline.


def noop(dictrows):
    """ No-op component

    Useful for pipeline component factories.
    """
    return dictrows


def remove_keys_with_empty_value(dictrows):
    for row in dictrows:
        yield {
            k: v for k, v in row.items()
            if k and v not in ('', None, [])
        }


##############################################################################
# Pipeline component factories


def warn_keys_with_unknown_value_except_for(*keys):
    def component(dictrows):
        for row in dictrows:
            for k, v in row.items():
                if k not in keys and text(v).lower() == 'unknown':
                    logger.warn('unknown %r for %s' % (k, row.get('uuid', '<empty uuid>')))
            yield row

    return component


def skip_rows_missing_all_keys(*keys):
    def component(dictrows):
        for row in dictrows:
            if not any(key in row for key in keys):
                row['_skip'] = True
            yield row

    return component


def skip_rows_with_all_key_value(**kw):
    def component(dictrows):
        for row in dictrows:
            if all(row[k] == v if k in row else False for k, v in kw.items()):
                row['_skip'] = True
            yield row

    return component


def skip_rows_without_all_key_value(**kw):
    def component(dictrows):
        for row in dictrows:
            if not all(row[k] == v if k in row else False for k, v in kw.items()):
                row['_skip'] = True
            yield row

    return component


def remove_keys(*keys):
    def component(dictrows):
        for row in dictrows:
            for key in keys:
                row.pop(key, None)
            yield row

    return component


def skip_rows_with_all_falsey_value(*keys):
    def component(dictrows):
        for row in dictrows:
            if all(not row[key] if key in row else False for key in keys):
                row['_skip'] = True
            yield row

    return component


def add_attachments(docsdir):
    def component(dictrows):
        for row in dictrows:
            for attachment_property in IS_ATTACHMENT:
                filename = row.get(attachment_property, None)
                if filename is None:
                    continue
                try:
                    path = find_doc(docsdir, filename)
                    row[attachment_property] = attachment(path)
                except ValueError as e:
                    row['_errors'] = repr(e)
            yield row

    return component


##############################################################################
# Read input from spreadsheets
#
# Downloading a zipfile of xlsx files from Google Drive is most convenient
# but it's better to check tsv into git.


def read_single_sheet(path, name=None):
    """ Read an xlsx, csv or tsv from a zipfile or directory
    """
    from zipfile import ZipFile
    from . import xlreader

    if name is None:
        root, ext = os.path.splitext(path)
        stream = open(path, 'r')

        if ext == '.xlsx':
            return read_xl(stream)

        if ext == '.tsv':
            return read_csv(stream, dialect='excel-tab')

        if ext == '.csv':
            return read_csv(stream)

        if ext == '.json':
            return read_json(stream)

        raise ValueError('Unknown file extension for %r' % path)

    if path.endswith('.xlsx'):
        return cast_row_values(xlreader.DictReader(open(path, 'rb'), sheetname=name))

    if path.endswith('.zip'):
        zf = ZipFile(path)
        names = zf.namelist()

        if (name + '.xlsx') in names:
            stream = zf.open(name + '.xlsx', 'r')
            return read_xl(stream)

        if (name + '.tsv') in names:
            stream = io.TextIOWrapper(zf.open(name + '.tsv'), encoding='utf-8')
            return read_csv(stream, dialect='excel-tab')

        if (name + '.csv') in names:
            stream = io.TextIOWrapper(zf.open(name + '.csv'), encoding='utf-8')
            return read_csv(stream)

        if (name + '.json') in names:
            stream = io.TextIOWrapper(zf.open(name + '.json'), encoding='utf-8')
            return read_json(stream)

    if os.path.isdir(path):
        root = os.path.join(path, name)

        if os.path.exists(root + '.xlsx'):
            stream = open(root + '.xlsx', 'rb')
            return read_xl(stream)

        if os.path.exists(root + '.tsv'):
            stream = open(root + '.tsv', 'rU')
            return read_csv(stream, dialect='excel-tab')

        if os.path.exists(root + '.csv'):
            stream = open(root + '.csv', 'rU')
            return read_csv(stream)

        if os.path.exists(root + '.json'):
            stream = open(root + '.json', 'r')
            return read_json(stream)

    return []


def read_xl(stream):
    from . import xlreader
    return cast_row_values(xlreader.DictReader(stream))


def read_csv(stream, **kw):
    import csv
    return cast_row_values(csv.DictReader(stream, **kw))


def read_json(stream):
    import json
    obj = json.load(stream)
    if isinstance(obj, dict):
        return [obj]
    return obj


##############################################################################
# Posting json
#
# This would a one liner except for logging

def request_url(item_type, method):
    def component(rows):
        for row in rows:
            if method == 'POST':
                url = row['_url'] = '/' + item_type
                yield row
                continue

            if '@id' in row:
                url = row['@id']
                if not url.startswith('/'):
                    url = '/' + url
                row['_url'] = url
                yield row
                continue

            # XXX support for aliases
            for key in ['uuid', 'accession']:
                if key in row:
                    url = row['_url'] = '/' + row[key]
                    break
            else:
                row['_errors'] = ValueError('No key found. Need uuid or accession.')

            yield row

    return component

def make_request(testapp, item_type, method):
    json_method = getattr(testapp, method.lower() + '_json')

    def component(rows):
        for row in rows:
            if row.get('_skip') or row.get('_errors') or not row.get('_url'):
                continue

            # Keys with leading underscores are for communicating between
            # sections
            value = row['_value'] = {
                k: v for k, v in row.items() if not k.startswith('_') and not k.startswith('@')
            }

            url = row['_url']
            row['_response'] = json_method(url, value, status='*')

            yield row

    return component


##############################################################################
# Logging


def trim(value):
    """ Shorten excessively long fields in error log
    """
    if isinstance(value, dict):
        return {k: trim(v) for k, v in value.items()}
    if isinstance(value, list):
        return [trim(v) for v in value]
    if isinstance(value, basestring) and len(value) > 160:
        return value[:77] + '...' + value[-80:]
    return value


def pipeline_logger(item_type, phase):
    def component(rows):
        created = 0
        updated = 0
        errors = 0
        skipped = 0
        count = 0
        for index, row in enumerate(rows):
            row_number = index + 2  # header row
            count = index + 1
            res = row.get('_response')

            if res is None:
                _skip = row.get('_skip')
                _errors = row.get('_errors')
                if row.get('_skip'):
                    skipped += 1
                elif _errors:
                    errors += 1
                    logger.error('%s row %d: Error PROCESSING: %s\n%r\n' % (item_type, row_number, _errors, trim(row)))
                yield row
                continue

            url = row.get('_url')
            uuid = row.get('uuid')

            if res.status_int == 200:
                updated += 1
                logger.debug('UPDATED: %s' % url)

            if res.status_int == 201:
                created += 1
                logger.debug('CREATED: %s' % res.location)

            if res.status_int == 409:
                logger.error('CONFLICT: %r' % res.json['detail'])

            if res.status_int == 422:
                logger.error('VALIDATION FAILED: %r' % trim(res.json['errors']))

            if res.status_int // 100 == 4:
                errors += 1
                logger.error('%s row %d: %s (%s)\n%r\n' % (item_type, row_number, res.status, url, trim(row['_value'])))

            yield row

        loaded = created + updated
        logger.info('Loaded %d of %d %s (phase %s). CREATED: %d, UPDATED: %d, SKIPPED: %d, ERRORS: %d' % (
             loaded, count, item_type, phase, created, updated, skipped, errors))

    return component


##############################################################################
# Attachments


def find_doc(docsdir, filename):
    path = None
    for dirpath in docsdir:
        candidate = os.path.join(dirpath, filename)
        if not os.path.exists(candidate):
            continue
        if path is not None:
            msg = 'Duplicate filenames: %s, %s' % (path, candidate)
            raise ValueError(msg)
        path = candidate
    if path is None:
        raise ValueError('File not found: %s' % filename)
    return path


def attachment(path):
    """ Create an attachment upload object from a filename

    Embeds the attachment as a data url.
    """
    import magic
    import mimetypes
    from PIL import Image
    from base64 import b64encode

    filename = os.path.basename(path)
    mime_type, encoding = mimetypes.guess_type(path)
    major, minor = mime_type.split('/')
    detected_type = magic.from_file(path, mime=True)

    # XXX This validation logic should move server-side.
    if not (detected_type == mime_type or
            detected_type == 'text/plain' and major == 'text'):
        raise ValueError('Wrong extension for %s: %s' % (detected_type, filename))

    with open(path, 'rb') as stream:
        attach = {
            'download': filename,
            'type': mime_type,
            'href': 'data:%s;base64,%s' % (mime_type, b64encode(stream.read()).decode('ascii'))
        }

        if mime_type in ('application/pdf', 'text/plain', 'text/tab-separated-values', 'text/html'):
            # XXX Should use chardet to detect charset for text files here.
            return attach

        if major == 'image' and minor in ('png', 'jpeg', 'gif', 'tiff'):
            # XXX we should just convert our tiffs to pngs
            stream.seek(0, 0)
            im = Image.open(stream)
            im.verify()
            if im.format != minor.upper():
                msg = "Image file format %r does not match extension for %s"
                raise ValueError(msg % (im.format, filename))

            attach['width'], attach['height'] = im.size
            return attach

    raise ValueError("Unknown file type for %s" % filename)


##############################################################################
# Pipelines


def combine(source, pipeline):
    """ Construct a combined generator from a source and pipeline
    """
    return reduce(lambda x, y: y(x), pipeline, source)


def process(rows):
    """ Pull rows through the pipeline
    """
    for row in rows:
        pass


def get_pipeline(testapp, docsdir, test_only, item_type, phase=None, method=None):
    pipeline = [
        skip_rows_with_all_key_value(test='skip'),
        skip_rows_with_all_key_value(_test='skip'),
        skip_rows_with_all_falsey_value('test') if test_only else noop,
        skip_rows_with_all_falsey_value('_test') if test_only else noop,
        remove_keys_with_empty_value,
        skip_rows_missing_all_keys('uuid', 'accession', '@id', 'name'),
        remove_keys('schema_version'),
        add_attachments(docsdir),
    ]
    if phase == 1:
        method = 'POST'
        pipeline.extend(PHASE1_PIPELINES.get(item_type, []))
    elif phase == 2:
        method = 'PUT'
        pipeline.extend(PHASE2_PIPELINES.get(item_type, []))

    pipeline.extend([
        request_url(item_type, method),
        remove_keys('uuid') if method in ('PUT', 'PATCH') else noop,
        make_request(testapp, item_type, method),
        pipeline_logger(item_type, phase),
    ])
    return pipeline


# Additional pipeline sections for item types

PHASE1_PIPELINES = {
    'user': [
        remove_keys('lab', 'submits_for'),
    ],
}


##############################################################################
# Phase 2 pipelines
#
# A second pass is required to cope with reference cycles. Only rows with
# filtered out keys are updated.


PHASE2_PIPELINES = {
    'user': [
        skip_rows_missing_all_keys('lab', 'submits_for'),
    ],
}


def load_all(testapp, filename, docsdir, test=False):
    for item_type in ORDER:
        try:
            source = read_single_sheet(filename, item_type)
        except ValueError:
            logger.error('Opening %s %s failed.', filename, item_type)
            continue
        pipeline = get_pipeline(testapp, docsdir, test, item_type, phase=1)
        process(combine(source, pipeline))

    for item_type in ORDER:
        if item_type not in PHASE2_PIPELINES:
            continue
        try:
            source = read_single_sheet(filename, item_type)
        except ValueError:
            continue
        pipeline = get_pipeline(testapp, docsdir, test, item_type, phase=2)
        process(combine(source, pipeline))


def load_test_data(app):
    from webtest import TestApp
    environ = {
        'HTTP_ACCEPT': 'application/json',
        'REMOTE_USER': 'TEST',
    }
    testapp = TestApp(app, environ)

    from pkg_resources import resource_filename
    inserts = resource_filename('snowflakes', 'tests/data/inserts/')
    docsdir = [resource_filename('snowflakes', 'tests/data/documents/')]
    load_all(testapp, inserts, docsdir)

    # users are by default created with random passwords... that's hard to guess
    # since this is just test data I think everybody should have the password Steve

    db = app.registry['dbsession']
    from snovault.storage import User
    for count, user in enumerate(db.query(User).all()):
        user.password = "Steve"

    import transaction
    transaction.commit()

    print ("all %s users get the password Steve" % (count))




"""Test Base Queue File"""
from unittest import (
    TestCase,
    mock,
)

from ..queues.base_queue import (
    BASE_QUEUE_TYPE,
    BaseQueueClient,
    BaseQueueMeta,
    BaseQueue,
)


MOCK_TIME = 123456.654321


class TestBaseClient(TestCase):
    '''Test Base Queue Client Class'''
    queue_options = {}
    queue_name = 'base-test-queue'
    queue_type = BASE_QUEUE_TYPE
    def test_get_queue(self):
        '''Test get_queue'''
        client = BaseQueueClient(self.queue_options)
        queue = client.get_queue(self.queue_name, self.queue_type)
        self.assertIsInstance(queue, BaseQueue)

    def test_get_queue_bad_type(self):
        '''Test get_queue with bad queue type'''
        client = BaseQueueClient(self.queue_options)
        try:
            _ = client.get_queue(self.queue_name, 'FakeType')
        except ValueError:
            pass
        else:
            self.fail('ValueError expected')


class TestBaseQueueMeta(TestCase):
    '''Test Base Queue Meta Class'''
    # pylint: disable=too-many-public-methods
    members = [
        ('_base_id', int),
        ('_errors', list),
        ('_uuid_count', int),
        ('_worker_conns', dict),
        ('_worker_results', dict),
        ('queue_name', str),
        ('_added_count', int), # not tested
        ('_errors_count', int), # not tested
        ('_success_count', int), # not tested
    ]
    meth_func = [
        # Errors
        'add_errors',
        'has_errors',
        'pop_errors',
        # Worker
        '_get_blank_worker',
        'add_worker_conn',
        'get_worker_conns',
        'get_worker_conn_count',
        'update_worker_conn',
        'save_work_results',
        # Uuids
        'has_uuids',
        'update_uuid_count',
        'get_server_restarts', # not tested
        'update_errors_count', # not tested
        'update_success_count', # not tested
    ]

    @classmethod
    @mock.patch('time.time', mock.MagicMock(return_value=MOCK_TIME))
    def setUpClass(cls):
        cls.mock_time_us = int(MOCK_TIME * 1000000)
        cls.queue_name = 'test_base_queueb'
        cls.queue_meta = BaseQueueMeta(cls.queue_name)
        cls.standard_errors = ['e' + str(cnt) for cnt in range(10)]

    @classmethod
    def tearDownClass(cls):
        '''Clean up after running test class'''
        pass

    def setUp(self):
        '''Setup before each method call'''
        pass

    def tearDown(self):
        """Clean up after each method call"""
        # pylint: disable=protected-access
        self.queue_meta._errors = []
        self.queue_meta._worker_results = {}
        self.queue_meta._worker_conns = {}

    # private/protected
    def test_init_dir(self):
        '''Test SimpleUuidServer has expected function and variables'''
        dir_items = [
            item
            for item in dir(self.queue_meta)
            if item[0:2] != '__'
        ]
        member_names = [name for name, _ in self.members]
        for item in dir_items:
            if item in member_names:
                continue
            if item in self.meth_func:
                continue
            raise AssertionError('Member or method not being tested:%s' % item)
        for item, item_type in self.members:
            self.assertTrue(hasattr(self.queue_meta, item))
            self.assertIsInstance(
                getattr(self.queue_meta, item),
                item_type
            )
        for item in self.meth_func:
            self.assertTrue(hasattr(self.queue_meta, item))
            self.assertTrue(
                str(type(getattr(self.queue_meta, item))) in [
                    "<class 'method'>",
                    "<class 'function'>"
                ]
            )

    def test_init_basic(self):
        '''Test basic queue meta init'''
        # pylint: disable=protected-access
        self.assertEqual(self.mock_time_us, self.queue_meta._base_id)
        self.assertListEqual([], self.queue_meta._errors)
        self.assertEqual(0, self.queue_meta._uuid_count)
        self.assertDictEqual({}, self.queue_meta._worker_conns)
        self.assertDictEqual({}, self.queue_meta._worker_results)

    # Errors
    def test_add_errors(self):
        '''Test add_errors'''
        errors_to_add = self.standard_errors.copy()
        expected_result = len(errors_to_add)
        result = self.queue_meta.add_errors(errors_to_add)
        self.assertEqual(expected_result, result)
        errors_to_add.sort()
        # pylint: disable=protected-access
        self.queue_meta._errors.sort()
        self.assertListEqual(errors_to_add, self.queue_meta._errors)

    def test_add_errors_none(self):
        '''Test add_errors when no errors'''
        errors_to_add = []
        expected_result = len(errors_to_add)
        result = self.queue_meta.add_errors(errors_to_add)
        self.assertEqual(expected_result, result)
        errors_to_add.sort()
        # pylint: disable=protected-access
        self.queue_meta._errors.sort()
        self.assertListEqual(errors_to_add, self.queue_meta._errors)

    def test_has_errors(self):
        '''Test has_errors'''
        # pylint: disable=protected-access
        self.queue_meta._errors = self.standard_errors.copy()
        self.assertTrue(self.queue_meta.has_errors())

    def test_has_errors_none(self):
        '''Test has_errors when no errors'''
        # pylint: disable=protected-access
        self.assertFalse(self.queue_meta.has_errors())

    def test_pop_errors(self):
        '''Test pop_errors'''
        # pylint: disable=protected-access
        errors_to_add = self.standard_errors.copy()
        self.queue_meta._errors = errors_to_add
        result = self.queue_meta.pop_errors()
        self.standard_errors.sort()
        result.sort()
        self.assertListEqual(self.standard_errors, result)
        self.assertListEqual([], self.queue_meta._errors)

    # Workers
    def test_get_blank_worker(self):
        '''Test _get_blank_worker'''
        # pylint: disable=protected-access
        self.assertDictEqual(
            {
                'uuid_cnt': 0,
                'get_cnt': 0,
            },
            self.queue_meta._get_blank_worker()
        )

    def test_add_worker_conn(self):
        '''Test add_worker_conn'''
        worker_id = 'test-worker-id'
        self.queue_meta.add_worker_conn(worker_id)
        # pylint: disable=protected-access
        self.assertTrue(worker_id in self.queue_meta._worker_results)
        self.assertListEqual([], self.queue_meta._worker_results[worker_id])
        self.assertTrue(worker_id in self.queue_meta._worker_conns)
        self.assertDictEqual(
            {
                'uuid_cnt': 0,
                'get_cnt': 0,
            },
            self.queue_meta._worker_conns[worker_id]
        )

    def test_get_worker_conns(self):
        '''Test get_worker_conns'''
        # pylint: disable=protected-access
        self.assertEqual(
            id(self.queue_meta._worker_conns),
            id(self.queue_meta.get_worker_conns())
        )

    def test_get_worker_conn_count(self):
        '''Test get_worker_conn_count'''
        # pylint: disable=protected-access
        self.assertEqual(
            len(self.queue_meta._worker_conns),
            self.queue_meta.get_worker_conn_count()
        )
        self.queue_meta._worker_conns['fake-id'] = {}
        self.assertEqual(
            len(self.queue_meta._worker_conns),
            self.queue_meta.get_worker_conn_count()
        )

    def test_update_worker_conn(self):
        '''Test update_worker_conn'''
        # pylint: disable=protected-access
        self.assertEqual(
            id(self.queue_meta._worker_conns),
            id(self.queue_meta.get_worker_conns())
        )


class TestBaseQueue(TestCase):
    '''Test Base Queue Class'''
    # No explict tests for the following since they pass directly to qmeta
    #  if that changes then add a test here.
    # - has_errors, pop_errors
    # - add_worker_conn, get_worker_conns, get_worker_conn_count,
    # update_worker_conn, save_work_results
    # - has_uuids, update_uuid_count

    # pylint: disable=too-many-public-methods
    members = [
        ('max_value_size', int),
        ('queue_type', str),
        ('_uuids', list),
        ('_qmeta', BaseQueueMeta),
        ('queue_name', str),
    ]
    meth_func = [
        # Errors
        'add_errors',
        'has_errors',
        'pop_errors',
        # Worker
        'add_worker_conn',
        'get_worker_conns',
        'get_worker_conn_count',
        'update_worker_conn',
        'save_work_results',
        # Uuids
        'has_uuids',
        'update_uuid_count',
        '_get_uuid',
        'get_uuids',
        '_load_uuid',
        'load_uuids',
        'update_finished',
        'close_indexing', # not tested
        'get_server_restarts', # not tested
        'update_errors_count', # not tested
        'update_success_count', # not tested
    ]

    @classmethod
    @mock.patch('time.time', mock.MagicMock(return_value=MOCK_TIME))
    def setUpClass(cls):
        cls.queue_options = {}
        cls.queue_name = 'base-test-queue'
        cls.queue_type = BASE_QUEUE_TYPE
        cls.mock_time_us = int(MOCK_TIME * 1000000)
        cls.client = BaseQueueClient(cls.queue_options)
        cls.queue = cls.client.get_queue(cls.queue_name, cls.queue_type)
        cls.standard_errors = ['e' + str(cnt) for cnt in range(10)]

    @classmethod
    def tearDownClass(cls):
        '''Clean up after running test class'''
        pass

    def setUp(self):
        '''Setup before each method call'''
        pass

    def tearDown(self):
        """Clean up after each method call"""
        # pylint: disable=protected-access
        self.queue._qmeta._errors = []
        self.queue._qmeta._worker_results = {}
        self.queue._qmeta._worker_conns = {}
        self.queue._uuids = []

    # private/protected
    def test_init_dir(self):
        '''Test SimpleUuidServer has expected function and variables'''
        dir_items = [
            item
            for item in dir(self.queue)
            if item[0:2] != '__'
        ]
        member_names = [name for name, _ in self.members]
        for item in dir_items:
            if item in member_names:
                continue
            if item in self.meth_func:
                continue
            raise AssertionError('Member or method not being tested:%s' % item)
        for item, item_type in self.members:
            self.assertTrue(hasattr(self.queue, item))
            self.assertIsInstance(
                getattr(self.queue, item),
                item_type
            )
        for item in self.meth_func:
            self.assertTrue(hasattr(self.queue, item))
            self.assertTrue(
                str(type(getattr(self.queue, item))) in [
                    "<class 'method'>",
                    "<class 'function'>"
                ]
            )

    def test_init_basic(self):
        '''Test basic queue meta init'''
        # pylint: disable=protected-access
        self.assertListEqual(self.queue._uuids, [])
        self.assertIsInstance(self.queue._qmeta, BaseQueueMeta)

    # Errors
    def test_add_errors(self):
        '''Test add_errors'''
        worker_id = 'wrkid-a'
        errors_to_add = [
            {'uuid': cnt, 'msg': 'err' + str(cnt)}
            for cnt in range(1, 4)
        ]
        res_err_cnt = self.queue.add_errors(worker_id, errors_to_add)
        self.assertEqual(res_err_cnt, len(errors_to_add))
        exp_sorted_errors = sorted(errors_to_add, key=lambda k: k['uuid'])
        # pylint: disable=protected-access
        qmeta_errors = self.queue._qmeta._errors
        res_sorted_errors = sorted(qmeta_errors, key=lambda k: k['uuid'])
        self.assertListEqual(exp_sorted_errors, res_sorted_errors)

    # Uuids
    def test_get_uuid(self):
        '''Test _get_uuid'''
        # pylint: disable=protected-access
        uuids_to_add = ['a', 'b', 'c']
        self.queue._uuids = uuids_to_add.copy()
        res_uuid = self.queue._get_uuid()
        self.assertEqual(res_uuid, uuids_to_add[-1])
        self.assertEqual(len(self.queue._uuids), len(uuids_to_add) - 1)

    def test_get_uuid_none(self):
        '''Test _get_uuid when no uuids'''
        # pylint: disable=protected-access
        res_uuid = self.queue._get_uuid()
        self.assertIsNone(res_uuid)

    def test_get_uuids_allshort(self):
        '''Test get_uuids all uuids with short cut'''
        uuids_to_add = ['a', 'b', 'c', 'd', 'e']
        # pylint: disable=protected-access
        self.queue._uuids = uuids_to_add.copy()
        res_uuids = self.queue.get_uuids(-10)
        uuids_to_add.sort()
        res_uuids.sort()
        self.assertListEqual(uuids_to_add, res_uuids)

    def test_get_uuids_overcnt(self):
        '''Test get_uuids all uuids over count'''
        uuids_to_add = ['a', 'b', 'c', 'd', 'e']
        # pylint: disable=protected-access
        self.queue._uuids = uuids_to_add.copy()
        res_uuids = self.queue.get_uuids(len(uuids_to_add) + 5)
        uuids_to_add.sort()
        res_uuids.sort()
        self.assertListEqual(uuids_to_add, res_uuids)

    def test_get_uuids_undercnt(self):
        '''Test get_uuids uuids with count'''
        uuids_to_add = ['a', 'b', 'c', 'd', 'e']
        # pylint: disable=protected-access
        self.queue._uuids = uuids_to_add.copy()
        get_cnt = len(uuids_to_add) - 2
        res_uuids = self.queue.get_uuids(get_cnt)
        self.assertEqual(get_cnt, len(res_uuids))
        exp_uuids = uuids_to_add[-1 * get_cnt:]
        exp_uuids.sort()
        res_uuids.sort()
        self.assertListEqual(exp_uuids, res_uuids)

    def test_get_uuids_none(self):
        '''Test get_uuids when no uuids'''
        res_uuids = self.queue.get_uuids(5)
        self.assertListEqual(res_uuids, [])

    def test_load_uuid(self):
        '''Test _load_uuid'''
        uuid_to_add = 'a'
        # pylint: disable=protected-access
        self.queue._load_uuid(uuid_to_add)
        self.assertListEqual(self.queue._uuids, [uuid_to_add])

    def test_load_uuid_many(self):
        '''Test _load_uuid many uuids, order should be reversed'''
        uuids_to_add = ['a', 'b', 'c', 'd', 'e']
        # pylint: disable=protected-access
        for uuid_to_add in uuids_to_add:
            self.queue._load_uuid(uuid_to_add)
        uuids_to_add.reverse()
        self.assertListEqual(self.queue._uuids, uuids_to_add)

    def test_load_uuids(self):
        '''Test load_uuids'''
        uuids_to_add = ['a', 'b', 'c', 'd', 'e']
        res_bytes_added, res_failed_uuids = self.queue.load_uuids(uuids_to_add)
        self.assertEqual(len(uuids_to_add), res_bytes_added)
        self.assertListEqual(res_failed_uuids, [])
        uuids_to_add.reverse()
        # pylint: disable=protected-access
        self.assertListEqual(self.queue._uuids, uuids_to_add)

    def test_load_uuids_none(self):
        '''Test load_uuids with none'''
        uuids_to_add = []
        res_bytes_added, res_failed_uuids = self.queue.load_uuids(uuids_to_add)
        self.assertEqual(len(uuids_to_add), res_bytes_added)
        self.assertListEqual(res_failed_uuids, [])
        uuids_to_add.reverse()
        # pylint: disable=protected-access
        self.assertListEqual(self.queue._uuids, uuids_to_add)

    def test_update_finished(self):
        '''Test update_finished'''
        worker_id = 'wrkid-b'
        uuid_cnt = 5
        get_cnt = 2
        self.queue.add_worker_conn(worker_id)
        self.queue.update_worker_conn(worker_id, uuid_cnt, get_cnt)
        results = {'errors': [], 'successes': uuid_cnt}
        res_msg = self.queue.update_finished(worker_id, results)
        exp_msg = 'Okay'
        self.assertEqual(res_msg, exp_msg)
        updated_worker_conn = self.queue.get_worker_conns()[worker_id]
        self.assertEqual(updated_worker_conn['uuid_cnt'], 0)
        self.assertEqual(updated_worker_conn['get_cnt'], get_cnt)
        # pylint: disable=protected-access
        worker_id_results = self.queue._qmeta._worker_results[worker_id]
        self.assertEqual(1, len(worker_id_results))
        self.assertDictEqual(results, worker_id_results[0])

    def test_update_finished_nowrks(self):
        '''Test update_finished with no workers'''
        worker_id = 'wrkid-c'
        results = {}
        res_msg = self.queue.update_finished(worker_id, results)
        exp_msg = 'Worker id(%s) DNE.' % worker_id
        self.assertEqual(res_msg, exp_msg)
        self.assertDictEqual(self.queue.get_worker_conns(), {})
        # pylint: disable=protected-access
        self.assertDictEqual(self.queue._qmeta._worker_results, {})

    def test_update_finished_errs(self):
        '''Test update_finished with errs'''
        worker_id = 'wrkid-d'
        uuid_cnt = 5
        get_cnt = 4
        errors_cnt = uuid_cnt - 3
        errors = [
            {'uuid': cnt, 'msg': 'err' + str(cnt)}
            for cnt in range(1, errors_cnt + 1)
        ]
        self.queue.add_worker_conn(worker_id)
        self.queue.update_worker_conn(worker_id, uuid_cnt, get_cnt)
        results = {'errors': errors, 'successes': uuid_cnt - errors_cnt}
        res_msg = self.queue.update_finished(worker_id, results)
        exp_msg = 'Okay'
        self.assertEqual(res_msg, exp_msg)
        updated_worker_conn = self.queue.get_worker_conns()[worker_id]
        self.assertEqual(updated_worker_conn['uuid_cnt'], 0)
        self.assertEqual(updated_worker_conn['get_cnt'], get_cnt)
        # pylint: disable=protected-access
        worker_id_results = self.queue._qmeta._worker_results[worker_id]
        self.assertEqual(1, len(worker_id_results))
        self.assertDictEqual(results, worker_id_results[0])

    def test_update_finished_noclose(self):
        '''Test update_finished cannot close worker'''
        worker_id = 'wrkid-d'
        uuid_cnt = 5
        get_cnt = 4
        errors_cnt = uuid_cnt - 3
        errors = [
            {'uuid': cnt, 'msg': 'err' + str(cnt)}
            for cnt in range(1, errors_cnt + 1)
        ]
        self.queue.add_worker_conn(worker_id)
        self.queue.update_worker_conn(worker_id, uuid_cnt, get_cnt)
        results = {'errors': errors, 'successes': uuid_cnt}
        res_msg = self.queue.update_finished(worker_id, results)
        exp_msg = 'Queue Server cannot close worker (%s)' % worker_id
        self.assertEqual(res_msg, exp_msg)
        updated_worker_conn = self.queue.get_worker_conns()[worker_id]
        self.assertEqual(
            updated_worker_conn['uuid_cnt'],
            uuid_cnt - len(errors) - uuid_cnt,  # final uuid_cnt is successes
        )
        self.assertEqual(updated_worker_conn['get_cnt'], get_cnt)
        # pylint: disable=protected-access
        worker_id_results = self.queue._qmeta._worker_results[worker_id]
        self.assertEqual(1, len(worker_id_results))
        self.assertDictEqual(results, worker_id_results[0])



import pytest
from pytest_bdd import scenarios

pytestmark = [
    pytest.mark.bdd,
    pytest.mark.usefixtures('workbook', 'admin_user'),
]


scenarios(
    'forms.feature',
    'page.feature',
    strict_gherkin=False
)



from snovault import (
    Item,
    collection,
)

# Test class for keys


def includeme(config):
    config.scan(__name__)


@collection(
    'testing-keys',
    properties={
        'title': 'Test keys',
        'description': 'Testing. Testing. 1, 2, 3.',
    },
    unique_key='testing_alias',
)
class TestingKey(Item):
    item_type = 'testing_key'
    schema = {
        'type': 'object',
        'properties': {
            'name': {
                'type': 'string',
                'uniqueKey': True,
            },
            'alias': {
                'type': 'string',
                'uniqueKey': 'testing_alias',
            },
        }
    }



from snosearch.fields import ResponseField
from snovault.elasticsearch.create_mapping import TEXT_FIELDS
from snovault.elasticsearch.searches.interfaces import NON_SORTABLE


class NonSortableResponseField(ResponseField):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def render(self, *args, **kwargs):
        self.parent = kwargs.get('parent')
        return {
            NON_SORTABLE: TEXT_FIELDS
        }



from itertools import chain
from itertools import zip_longest
from elasticsearch.exceptions import (
    ConflictError,
    ConnectionError,
    NotFoundError,
    TransportError,
)
from pyramid.view import view_config
from pyramid.settings import asbool
from sqlalchemy.exc import StatementError
from snovault import (
    COLLECTIONS,
    DBSESSION,
    STORAGE
)
from snovault.storage import (
    TransactionRecord,
)
from snovault.util import get_uuids_from_file
from urllib3.exceptions import ReadTimeoutError
from .interfaces import (
    ELASTIC_SEARCH,
    INDEXER
)
import datetime
import logging
import pytz
import time
import copy
import json
import requests
import re
import boto3
import socket


AWS_REGION = 'us-west-2'
_HOSTNAME = socket.gethostname()
SEARCH_MAX = 99999  # OutOfMemoryError if too high
HEAD_NODE_INDEX = 'head_node'
INDEXING_NODE_INDEX = 'indexing_node'


es_logger = logging.getLogger("elasticsearch")
es_logger.setLevel(logging.ERROR)
log = logging.getLogger('snovault.elasticsearch.es_index_listener')

def includeme(config):
    config.add_route('_indexer_state', '/_indexer_state')
    config.scan(__name__)


def _get_this_instance_name(instance_name=None):
    ec2 = boto3.client('ec2', region_name=AWS_REGION)
    if not instance_name:
        hostname = _HOSTNAME.replace('ip-', '').replace('-', '.')
        response = ec2.describe_instances(Filters=[
                {'Name': 'private-ip-address', 'Values': [hostname]},
        ])
    else:
        response = ec2.describe_instances(Filters=[
            {'Name': 'tag:Name', 'Values': [instance_name]},
        ])
    if not response.get('ResponseMetadata', {}).get('HTTPStatusCode') == 200:
        return None 
    instance_name = None
    instance_state = None
    for reservation in response.get('Reservations', []):
        for instance in reservation.get('Instances', []):
            instance_state = instance['State']['Name']
            for tag_obj in instance.get('Tags', []):
                if tag_obj.get('Key') == 'Name':
                    instance_name = tag_obj.get('Value')
                    break
    return instance_name, instance_state


def setup_indexing_nodes(request, indexer_state, reset=False):
    time_now = float(time.time())
    did_fail = False
    is_indexing_node = False
    node_state_template = {
        'node_index': None,
        'instance_name': None,
        'instance_state': None,
        'waiting_on_remote': False,
        'started_indexing': False,
        'done_indexing': False,
        'last_run_time': str(time_now),
    }
    # Get/Setup head node state in local elasticsearch meta data
    head_node = indexer_state.get_obj(HEAD_NODE_INDEX)
    indexing_node = indexer_state.get_obj(INDEXING_NODE_INDEX)
    indexer_tag = '-indexer'
    # Check which node this is
    instance_name, instance_state = _get_this_instance_name()
    if not instance_name:
        did_fail = True
        return head_node, indexing_node, time_now, did_fail, is_indexing_node
    elif not instance_name.replace(indexer_tag, '') == instance_name:
        # indexing node does not manage states
        is_indexing_node = True
        return head_node, indexing_node, time_now, did_fail, is_indexing_node
    # Handle Setup or Refresh
    if not head_node or not indexing_node or reset:
        # Setup node state in elasticsearch
        # Save this node as head node
        head_node = copy.deepcopy(node_state_template)
        head_node['node_index'] = HEAD_NODE_INDEX
        head_node['instance_name'] = instance_name
        head_node['instance_state'] = instance_state
        indexer_state.put_obj(head_node['node_index'], head_node)
        # Create indexing node
        indexer_name = f"{instance_name}{indexer_tag}"
        indexer_instance_name, indexer_instance_state = _get_this_instance_name(
            instance_name=indexer_name
        )
        if not indexer_instance_name:
            did_fail = True
            return head_node, indexing_node, time_now, did_fail, is_indexing_node
        indexing_node = copy.deepcopy(node_state_template)
        indexing_node['node_index'] = INDEXING_NODE_INDEX
        indexing_node['instance_name'] = indexer_instance_name
        indexing_node['instance_state'] = indexer_instance_state
        indexing_node['last_run_time'] = str(float(time.time()))
        indexer_state.put_obj(indexing_node['node_index'], indexing_node)
    else:
        # Refresh aws instance state
        # Head
        instance_name, instance_state = _get_this_instance_name(
            instance_name=head_node['instance_name']
        )
        if instance_name and instance_state:
            head_node['instance_state'] = instance_state
            indexer_state.put_obj(head_node['node_index'], head_node)
        else:
            log.warning('Remote indexer failed to update head node')
        # Indexer
        instance_name, instance_state = _get_this_instance_name(
            instance_name=indexing_node['instance_name']
        )
        if instance_name and instance_state:
            indexing_node['instance_state'] = instance_state
            if instance_state == 'stopped':
                # reset clock if stopped
                indexing_node['last_run_time'] = str(float(time.time()))
            indexer_state.put_obj(indexing_node['node_index'], indexing_node)
        else:
            log.warning('Remote indexer failed to update indexer node')
    return head_node, indexing_node, time_now, did_fail, is_indexing_node


class IndexerState(object):
    _is_reindex_base = '_is_reindex'
    # Keeps track of uuids and indexer state by cycle.  Also handles handoff of uuids to followup indexer
    def __init__(self, es, index, title='primary', followups=[]):
        self.es = es
        self.index = index  # "index where indexerstate is stored"

        self.title           = title
        self.state_id        = self.title + '_indexer'       # State of the current or last cycle
        self.todo_set        = self.title + '_in_progress'   # one cycle of uuids, sent to the Secondary Indexer
        #self.failed_set      = self.title + '_failed'
        #self.done_set        = self.title + '_done'          # Trying to get all uuids from 'todo' to this set
        self.troubled_set    = self.title + '_troubled'      # uuids that failed to index in any cycle
        self.last_set        = self.title + '_last_cycle'    # uuids in the most recent finished cycle
        self.success_set     = None                          # None is the same as self.done_set
        self.cleanup_this_cycle = [self.todo_set]  # ,self.failed_set,self.done_set]  # Clean up at end of current cycle
        self.cleanup_last_cycle = [self.last_set,self.troubled_set]              # Clean up at beginning of next cycle
        self.override           = 'reindex_' + self.title      # If exists then reindex all
        # DO NOT INHERIT! These keys are for passing on to other indexers
        self.followup_prep_list      = 'primary_followup_prep_list' # Setting up the uuids to be handled by a followup process
        self.staged_for_vis_list     = 'staged_for_vis_indexer' # Followup list is added to here to pass baton
        self.staged_for_regions_list = 'staged_for_region_indexer'     # Followup list is added to here to pass baton
        self.followup_lists = []                                     # filled dynamically
        for name in followups:
            if name != '':
                list_id = 'staged_for_' + name
                assert list_id == self.staged_for_vis_list or list_id == self.staged_for_regions_list
                self.followup_lists.append(list_id)
        self.clock = {}
        self._is_reindex_key = self.title + self._is_reindex_base
        self.is_reindexing = False
        # Initial indexing will also be true if the indexing has been reset
        self.is_initial_indexing = False
        # some goals:
        # 1) Detect and recover from interrupted cycle - working but ignored for now
        # 2) Record (double?) failures and consider blacklisting them - not tried, could do.

    def _del_is_reindex(self):
        # Flag should not be cleared until finish_cycle function is called
        return self.delete_objs([self._is_reindex_key])

    def _get_is_reindex(self):
        obj = self.get_obj(self._is_reindex_key)
        return obj.get('is_reindex') is True

    def _set_is_reindex(self):
        # Flag should be set in request_reindex funciton
        self.put_obj(self._is_reindex_key, {'is_reindex': True})

    def log_reindex_init_state(self):
        # Must call after priority cycle
        if self.is_reindexing and self.is_initial_indexing:
            log.info('%s is reindexing all', self.title)
        elif self.is_reindexing:
            log.info('%s is reindexing', self.title)
        elif self.is_initial_indexing:
            log.info('%s is initially indexing', self.title)

    # Private-ish primitives...
    def get_obj(self, id, doc_type='meta'):
        try:
            return self.es.get(index=self.index, doc_type=doc_type, id=id).get('_source',{})  # TODO: snovault/meta
        except:
            return {}

    def put_obj(self, id, obj, doc_type='meta'):
        try:
            self.es.index(index=self.index, doc_type=doc_type, id=id, body=obj)
        except ConnectionError as ecp:
            log.warn(f"Failed to save {id} to es due ConnectionError")
        except Exception as ecp:
            ecp_name = ecp.__class__.__name__
            log.warn(f"Failed to save {id} to es due to exception {ecp_name}: {repr(ecp)}")

    def delete_objs(self, ids, doc_type='meta'):
        for id in ids:
            try:
                self.es.delete(index=self.index, doc_type=doc_type, id=id)
            except:
                pass

    def get_list(self, id):
        return self.get_obj(id).get('list',[])

    def get_count(self, id):
        return self.get_obj(id).get('count',0)

    def put_list(self, id, a_list):
        return self.put_obj(id, { 'list': a_list, 'count': len(a_list) })

    #def get_diff(self,orig_id, subtract_ids):
    #    result_set = set(self.get_list(orig_id))
    #
    #    if len(result_set) > 0:
    #        for id in subtract_ids:
    #            subtract_list = self.get_list(id)
    #            if len(subtract_list):
    #                result_set = result_set.difference(set(subtract_list))
    #    return result_set

    def set_add(self, id, vals):
        set_to_update = set(self.get_list(id))
        if len(set_to_update) > 0:
            set_to_update.update(vals)
        else:
            set_to_update = set(vals)
        self.put_list(id, set_to_update)

    def list_extend(self, id, vals):
        list_to_extend = self.get_list(id)
        if len(list_to_extend) > 0:
            list_to_extend.extend(vals)  # TODO: consider capping at SEARCH_MAX (keeping count but not uuids).  Requires followup handoff work.
        else:
            list_to_extend = vals

        self.put_list(id, list_to_extend)

    def rename_objs(self, from_id, to_id):
        val = self.get_list(from_id)
        if val:
            self.put_list(to_id, val)
            self.delete_objs([from_id])

    # Public access...
    def get(self):
        '''Returns the basic state info'''
        return self.get_obj(self.state_id)

    def put(self, state):
        '''Update the basic state info'''
        # Don't save errors in es
        errors = state.pop('errors', None)

        state['title'] = self.state_id
        self.put_obj(self.state_id, state)

        if errors is not None:
            state['errors'] = errors

    def request_reindex(self,requested):
        '''Requests full reindexing on next cycle'''
        if requested == 'all':
            self._set_is_reindex()
            if self.title == 'primary':  # If primary indexer delete the original master obj
                self.delete_objs(["indexing"])  # http://localhost:9201/snovault/meta/indexing
            else:
                self.put_obj(self.override, {self.title : 'reindex', 'all_uuids': True})

        else:
            uuid_list = requested.split(',')
            uuids = set()
            while uuid_list:
                uuid = uuid_list.pop(0)
                if len(uuid) > 0 and re.match("[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}", uuid):
                    uuids.add(uuid)
                else:
                    return "Requesting reindex of at least one uninterpretable uuid: '%s'" % (uuid)
            override_obj = self.get_obj(self.override)
            if 'uuids' not in override_obj.keys():
                override_obj['uuids'] = list(uuids)
            else:
                uuids |= set(override_obj['uuids'])
                override_obj['uuids'] = list(uuids)
            self._set_is_reindex()
            self.put_obj(self.override, override_obj)
        return None

    def all_indexable_uuids(self, request):
        '''returns list of uuids pertinant to this indexer.'''
        return list(all_uuids(request.registry))

    def reindex_requested(self, request):
        '''returns list of uuids if a reindex was requested.'''
        override = self.get_obj(self.override)
        if override:
            if override.get('all_uuids', False):
                self.delete_objs([self.override] + self.followup_lists)
                return self.all_indexable_uuids(request)
            else:
                uuids =  override.get('uuids',[])
                uuid_count = len(uuids)
                if uuid_count > 0:
                    if uuid_count > SEARCH_MAX:
                        self.delete_objs([self.override] + self.followup_lists)
                    else:
                        self.delete_objs([self.override])
                    return uuids
        return None


    def get_initial_state(self):
        '''Useful to initialize at idle cycle'''
        new_state = { 'title': self.state_id, 'status': 'idle'}
        state = self.get()
        for var in ['cycles']:  # could expand this list
            val = state.pop(var,None)
            if val is not None:
                new_state[var] = val
        # Make sure indexer is registered:
        self.set_add("registered_indexers", [self.state_id])
        return new_state

    def start_clock(self, name):
        '''Can start a named clock and use it later to figure out elapsed time'''
        self.clock[name] = datetime.datetime.now(pytz.utc)

    def elapsed(self, name):
        '''Returns string of time elapsed since named clock started.'''
        start = self.clock.get(name)
        if start is None:
            return 'unknown'
        else:
            return str(datetime.datetime.now(pytz.utc) - start)

    def priority_cycle(self, request):
        '''Initial startup, reindex, or interupted prior cycle can all lead to a priority cycle.
           returns (discovered xmin, uuids, whether previous cycle was interupted).'''
        # Not yet started?
        initialized = self.get_obj("indexing")  # http://localhost:9201/snovault/meta/indexing
        self.is_reindexing = self._get_is_reindex()
        if not initialized:
            self.is_initial_indexing = True
            self.delete_objs([self.override] + self.followup_lists)
            state = self.get()
            state['status'] = 'uninitialized'
            self.put(state)
            return (-1, [], False)  # primary indexer will know what to do and secondary indexer should do nothing yet

        state = self.get()

        # Rare call for reindexing...
        reindex_uuids = self.reindex_requested(request)
        if reindex_uuids is not None and reindex_uuids != []:
            uuids_count = len(reindex_uuids)
            log.warn('%s reindex of %d uuids requested' % (self.state_id, uuids_count))
            return (-1, reindex_uuids, False)

        if state.get('status', '') != 'indexing':
            return (-1, [], False)

        xmin = state.get('xmin', -1)
        #snapshot = state.get('snapshot', None)
        if xmin == -1:  # or snapshot is None:
            return (-1, [], False)

        #assert(self.get_count(self.done_set) == 0)  # Valid for cycle-level accounting only
        #undone_uuids = self.get_diff(self.todo_set, [self.done_set])  # works for any accountingu
        undone_uuids = self.get_list(self.todo_set)                    # works fastest for cycle-level accounting
        if len(undone_uuids) <= 0:  # TODO SEARCH_MAX?  SEARCH_MAX/10
            return (-1, [], False)

        # Note: do not clean up last cycle yet because we could be restarted multiple times.
        return (xmin, undone_uuids, True)


    def prep_for_followup(self, xmin, uuids):
        '''Prepare a cycle of uuids for passing to a followup indexer (e.g. audits, viscache)'''
        prep_list = [ "xmin:%s" % xmin ]
        prep_list.extend(uuids)
        self.put_list(self.followup_prep_list, prep_list)
        # No need to preserve anything on the prep_list as it passes to the staged list in one cycle.

    def start_cycle(self, uuids, state=None):
        '''Every indexing cycle must be properly opened.'''
        self.clock = {}
        self.start_clock('cycle')
        if state is None:
            state = self.get()
        state['cycle_started'] = datetime.datetime.now().isoformat()
        state['status'] = 'indexing'
        state['cycle_count'] = len(uuids)

        self.put(state)
        self.delete_objs(self.cleanup_last_cycle)
        self.delete_objs(self.cleanup_this_cycle)
        self.put_list(self.todo_set, set(uuids))
        return state

    def add_errors(self, errors, finished=True):
        '''To avoid 16 worker concurency issues, errors are recorded at the end of a cycle.'''
        uuids = [err['uuid'] for err in errors]  # better be uuids!
        if len(uuids) > 0:
            # Forget extend, this is cycle-level accounting
            #self.list_extend(self.failed_set, uuids)
            # Forget failed_set, this is cycle-level accounting
            #self.put_list(self, self.failed_set, uuids)

            if finished:
                # Forget sets... cycle-level accounting so errors => failed_set => troubled_set all in one cycle.
                # handle troubled uuids:
                #troubled_uuids = set(self.get_list(self.failed_set))
                #if len(troubled_uuids):
                #    self.set_add(self.troubled_set, troubled_uuids)
                #    # TODO: could make doubled_troubled set and use it to blacklist uuids
                self.put_list(self.troubled_set, uuids)

    def finish_cycle(self, state, errors=None):
        '''Every indexing cycle must be properly closed.'''

        if errors:  # By handling here, we avoid overhead and concurrency issues of uuid-level accounting
            self.add_errors(errors)

        # pass any staged items to followup
        if self.followup_prep_list is not None:
            # TODO: send signal for 'all' when appropriate.  Saves the following expensive lines.
            hand_off_list = self.get_list(self.followup_prep_list)
            if len(hand_off_list) > 0:  # Have to push because ready_list may still have previous cycles in it
                for id in self.followup_lists:
                    self.list_extend(id, hand_off_list)
                    #log.warn("prmary added to %s" % id)
                self.delete_objs([self.followup_prep_list])

        # cycle-level accounting so todo => done => last in this function
        #self.rename_objs(self.todo_set, self.done_set)
        done_count = self.get_count(self.todo_set)
        self.rename_objs(self.todo_set, self.last_set)

        if self.success_set is not None:
            state[self.title + '_updated'] = self.get_count(self.success_set)
        cycle_count = state.pop('cycle_count', None)
        #assert(cycle_count == done_count)
        state['indexed'] = done_count

        #self.rename_objs(self.done_set, self.last_set)   # cycle-level accounting so todo => done => last in this function
        self.delete_objs(self.cleanup_this_cycle)
        state['status'] = 'done'
        state['cycles'] = state.get('cycles', 0) + 1
        state['cycle_took'] = self.elapsed('cycle')

        self.put(state)
        self._del_is_reindex()
        return state

    def get_notice_user(self, user, bot_token):
        '''Returns the user token for a named user.'''
        slack_users = self.get_obj('slack_users',{})
        if not slack_users:
            try:
                r = requests.get('https://slack.com/api/users.list?token=%s' % (bot_token))
                resp = json.loads(r.text)
                if not resp['ok']:  # and resp.get('error','') != 'ratelimited':
                    log.warn(resp)  # too many at once: {'error': 'ratelimited', 'ok': False}
                    return None
                members = resp.get('members',[])
                for member in members:
                    slack_users[member['name']] = member['id']
                self.put_obj('slack_users',slack_users)
            except:
                return None
        return slack_users.get(user)

    def set_notices(self, from_host, who=None, bot_token=None, which=None):
        '''Set up notification so that slack bot can send a message when indexer finishes'''
        if who is None and bot_token is None:
            return "ERROR: must specify who to notify or bot_token"
        if which is None:
            which = self.state_id
        if which == 'all':
            which = 'all_indexers' # needed because of elasticsearch conflicts with 'all'
        elif which not in self.get_list("registered_indexers"):
            if which + '_indexer' in self.get_list("registered_indexers"):
                which += '_indexer'
            else:
                return "ERROR: unknown indexer to monitor: %s" % (which)

        notify = self.get_obj('notify', 'default')
        if bot_token is not None:
            notify['bot_token'] = bot_token

        if 'from' not in notify:
            notify['from'] = from_host

        user_warns = ''
        if who is not None:
            if who in ['none','noone','nobody','stop','']:
                notify.pop(which, None)
            else:
                indexer_notices = notify.get(which,{})
                if which == 'all_indexers':  # each indexer will have to finish
                    if 'indexers' not in indexer_notices:
                        indexer_notices['indexers'] = self.get_list("registered_indexers")
                users = who.split(',')
                if 'bot_token' in notify:
                    for name in users:
                        if self.get_notice_user(name,notify['bot_token']) is None:
                            user_warns += ', ' + name
                who_all = indexer_notices.get('who',[])
                who_all.extend(users)
                indexer_notices['who'] = list(set(who_all))
                notify[which] = indexer_notices
                # either self.state_id: {who: [...]} or 'all_indexers': {'indexers': [...], 'who': [...]}

        self.put_obj('notify', notify, 'default')
        if user_warns != '':
            user_warns = 'Unknown users: ' + user_warns[2:]
        if 'bot_token' not in notify:
            return "WARNING: bot_token is required. " + user_warns
        elif user_warns != '':
            return "WARNING: " + user_warns

        return None

    def get_notices(self, full=False):
        '''Get the notifications'''
        notify = self.get_obj('notify','default')
        if full:
            return notify
        notify.pop('bot_token', None)
        notify.pop('from', None)
        for which in notify.keys():
            if len(notify[which].get('who',[])) == 1:
                notify[which] = notify[which]['who'][0]
        indexers = self.get_list("registered_indexers")
        if self.title == 'primary':  # Primary will show all notices
            indexers.append('all_indexers')
            for indexer in indexers:
                if notify.get(indexer,{}):
                    return notify  # return if anything
        else:  # non-primary will show specific notice and all
            indexers.remove(self.state_id)
            for indexer in indexers:
                notify.pop(indexer,None)
            for indexer in [self.state_id, 'all_indexers']:
                if notify.get(indexer,{}):
                    return notify  # return if anything
        return {}

    def send_notices(self):
        '''Sends notifications when indexer is done.'''
        # https://slack.com/api/chat.postMessage?token=xoxb-1974789...&channel=U1KPQK1HN&text=Yay!
        notify = self.get_obj('notify','default')
        if not notify:
            return
        if 'bot_token' not in notify or 'from' not in notify:
            return  # silent failure, but leaves notify unchanged for evidence

        changed = False
        text = None
        who = []
        if 'all_indexers' in notify:  # 'all_indexers': {'indexers': [...], 'who': [...]}
            # if all indexers are done, then report
            indexers = notify['all_indexers'].get('indexers',[])
            if self.state_id in indexers:
                # Primary must finish before follow up indexers can remove themselves from list
                if self.state_id == 'primary_indexer' or 'primary_indexer' not in indexers:
                    indexers.remove(self.state_id)
                    if len(indexers) > 0:
                        notify['all_indexers']['indexers'] = indexers
                        changed = True
            if len(indexers) == 0:
                who.extend(notify['all_indexers'].get('who',[]))
                notify.pop('all_indexers',None)
                changed = True
                text='All indexers are done for %s/_indexer_state' % (notify['from'])
        if self.state_id in notify:  # self.state_id: {who: [...]}
            who.extend(notify[self.state_id].get('who',[]))
            notify.pop(self.state_id, None)
            changed = True
            if text is None:
                text='%s is done for %s' % (self.state_id, notify['from'])
                if self.title == 'primary':
                    text += '/_indexer_state'
                else:
                    text += '/_%sindexer_state' % (self.title)
        if len(who) > 0 and text is not None:
            who = list(set(who))
            users = ''
            msg = ''
            if len(who) == 1:
                #channel='U1KPQK1HN'  # TODO: look up user token
                channel = self.get_notice_user(who[0],notify['bot_token'])
                if channel:
                    msg = 'token=%s&channel=%s&text=%s' % (notify['bot_token'],channel,text)
                # otherwise fall back on generic message.
            if msg == '':  # This will catch multiple users AND single users for which a channel could not be found
                channel='dcc-private'
                for user in who:
                    users += '@' + user + ', '
                msg = 'token=%s&channel=%s&link_names=true&text=%s%s' % (notify['bot_token'],channel,users,text)

            try:
                r = requests.get('https://slack.com/api/chat.postMessage?' + msg)
                resp = json.loads(r.text)
                if not resp['ok']:
                    log.warn(resp)
            except:
                log.warn("Failed to notify via slack: [%s]" % (msg))

        if changed:  # alter notify even if error, so the same error doesn't flood log.
            self.put_obj('notify', notify, 'default')

    def display(self, uuids=None):
        display = {}
        display['state'] = self.get()
        if display['state'].get('status','') == 'indexing' and 'cycle_started' in display['state']:
            started = datetime.datetime.strptime(display['state']['cycle_started'],'%Y-%m-%dT%H:%M:%S.%f')
            display['state']['indexing_elapsed'] = str(datetime.datetime.now() - started)
        display['title'] = display['state'].get('title',self.state_id)
        display['uuids_in_progress'] = self.get_count(self.todo_set)
        display['uuids_troubled'] = self.get_count(self.troubled_set)
        display['uuids_last_cycle'] = self.get_count(self.last_set)
        if self.followup_prep_list is not None:
            display['to_be_staged_for_follow_up_indexers'] = self.get_count(self.followup_prep_list)
        if self.title == 'primary':
            for id in self.followup_lists:
                display[id] = self.get_count(id)
        else:
            id = 'staged_for_%s_list' % (self.title)
            display['staged_by_primary'] = self.get_count(id)

        reindex = self.get_obj(self.override)
        if reindex:
            uuids = reindex.get('uuids')
            if uuids is not None:
                display['reindex_requested'] = uuids
            elif reindex.get('all_uuids',False):
                display['reindex_requested'] = 'all'
        notify = self.get_notices()
        if notify:
            display['notify_requested'] = notify
        display['now'] = datetime.datetime.now().isoformat()

        if uuids is not None:
            uuids_to_show = []
            uuid_list = self.get_obj(self.todo_set)
            if not uuid_list:
                uuids_to_show = 'No uuids indexing'
            else:
                uuid_start = 0
                try:
                    uuid_start = int(uuids)
                except:
                    pass
                if uuid_start < uuid_list.get('count',0):
                    uuid_end = uuid_start+100
                    if uuid_start > 0:
                        uuids_to_show.append("... skipped first %d uuids" % (uuid_start))
                    uuids_to_show.extend(uuid_list['list'][uuid_start:uuid_end])
                    if uuid_list.get('count',0) > uuid_end:
                        uuids_to_show.append("another %d uuids..." % (uuid_list.get('count',0) - uuid_end))
                elif uuid_start > 0:
                    uuids_to_show.append("skipped past all %d uuids" % (uuid_list.get('count',0)))
                else:
                    uuids_to_show = 'No uuids indexing'
            display['uuids_in_progress'] = uuids_to_show

        return display

@view_config(route_name='_indexer_state', request_method='GET', permission="index")
def indexer_state_show(request):
    es = request.registry[ELASTIC_SEARCH]
    INDEX = request.registry.settings['snovault.elasticsearch.index']
    state = IndexerState(es,INDEX)

    # requesting reindex
    reindex = request.params.get("reindex")
    if reindex is not None:
        msg = state.request_reindex(reindex)
        if msg is not None:
            return msg

    # requesting reset on remote indexing
    reset_remote = request.params.get("reset_remote")
    if reset_remote is not None:
        setup_indexing_nodes(request, state, reset=True)
        return 'did reset indexing nodes'

    # Requested notification
    who = request.params.get("notify")
    bot_token = request.params.get("bot_token")
    if who is not None or bot_token is not None:
        notices = state.set_notices(request.host_url, who, bot_token, request.params.get("which"))
        if notices is not None:
            return notices

    display = state.display(uuids=request.params.get("uuids"))
    # getting count is complicated in es5 as docs are separate indexes
    item_types = all_types(request.registry)
    count = 0
    for item_type in item_types:
        # TODO: index list should be replaced by index alias:
        #       https://www.elastic.co/guide/en/elasticsearch/reference/6.2/indices-aliases.html
        try:
            type_count = es.count(index=item_type).get('count',0)
            count += type_count
        except:
            pass
    if count:
        display['docs_in_index'] = count
    else:
        display['docs_in_index'] = 'Not Found'

    if not request.registry.settings.get('testing',False):  # NOTE: _indexer not working on local instances
        try:
            r = requests.get('http://localhost/_indexer')
            display['listener'] = json.loads(r.text)
            display['status'] = display['listener']['status']
        except:
            log.error('Error getting /_indexer', exc_info=True)

    display['registered_indexers'] = state.get_list('registered_indexers')
    # always return raw json
    request.query_string = "format=json"
    return display


def all_types(registry):
    collections = registry[COLLECTIONS]
    return sorted(collections.by_item_type)


def heterogeneous_stream(generator_map):
    '''
    Will zip together generators and yield until all are exhausted, e.g.:

    >>> generator_map = {
        'experiments': (e for e in experiment_uuids),
        'files': (f for f in file_uuids)
    }
    >>> assert list(heterogeneous_stream(generator_map)) == [e1, f1, e2, f2, ..., eN, fM]

    where:
         N is number of experiment uuids
         M is number of file uuids
         N doesn't have to equal M

    This allows for structured mixing of collections, though the stream will become
    more homogeneous as shorter collections are exhausted.
    '''
    for x in chain(*zip_longest(*generator_map.values())):
        if x is None:
            continue
        yield x


def all_uuids(registry, types=None):
    # First index user and access_key so people can log in
    collections = registry[COLLECTIONS]
    initial = ['user', 'access_key']
    for collection_name in initial:
        collection = collections.by_item_type.get(collection_name, [])
        # for snovault test application, there are no users or keys
        if types is not None and collection_name not in types:
            continue
        for uuid in collection:
            yield str(uuid)
    if asbool(registry.settings.get('use_small_db', False)):
        for uuid in get_uuids_from_file(registry.settings.get('small_db_path')):
            yield uuid
        return
    uuid_generator_map = {}
    for collection_name in sorted(collections.by_item_type):
        if collection_name in initial:
            continue
        if types is not None and collection_name not in types:
            continue
        collection = collections.by_item_type[collection_name]
        uuid_generator_map[collection_name] = (
            uuid
            for uuid in collection
        )
    for uuid in heterogeneous_stream(uuid_generator_map):
        yield str(uuid)



import pytest
from unittest.mock import Mock


class TestPickStorage(object):

    def test_get_rev_links_with_switched_storage(self):
        # If we have a model that was loaded from elasticsearch
        # but the datastore is now database, we need to be sure
        # to get the rev links from the database.
        from pyramid.testing import (
            DummyRequest,
            testConfig,
        )
        from ..esstorage import (
            CachedModel,
            PickStorage,
        )

        model = CachedModel({'_source': {'uuid': 'dummy'}})
        model2 = Mock()
        read = Mock()
        write = Mock()
        write.get_by_uuid.return_value = model2
        storage = PickStorage(read, write)

        request = DummyRequest()
        request.datastore = 'database'
        with testConfig(request=request):
            storage.get_rev_links(model, 'rel')

        assert write.get_rev_links.called_once
        rev_links_model = write.get_rev_links.call_args[0][0]
        assert rev_links_model is model2


def test_pick_storage_get_by_unique_key_forces_database_request_when_model_is_invalidated(dummy_request):
    from pyramid.testing import testConfig
    from ..esstorage import PickStorage
    # Uses ES when model not invalidated.
    dummy_request.datastore = 'elasticsearch'
    es_model = Mock()
    pg_model = Mock()
    read = Mock()
    write = Mock()
    read.get_by_unique_key.return_value = es_model
    write.get_by_unique_key.return_value = pg_model
    es_model.invalidated.return_value = False
    pg_model.invalidated.return_value = False
    storage = PickStorage(read, write)
    with testConfig(request=dummy_request):
        model = storage.get_by_unique_key('accession', 'SNOFL000LSQ')
    assert model is es_model
    assert dummy_request.datastore == 'elasticsearch'
    # Switches to PG when model invalidated.
    es_model.invalidated.return_value = True
    with testConfig(request=dummy_request):
        model = storage.get_by_unique_key('accession', 'SNOFL000LSQ')
    assert model is pg_model
    assert dummy_request.datastore == 'database'


def test_pick_storage_get_by_uuid_forces_database_request_when_model_is_invalidated(dummy_request):
    from pyramid.testing import testConfig
    from ..esstorage import PickStorage
    # Uses ES when model not invalidated.
    dummy_request.datastore = 'elasticsearch'
    es_model = Mock()
    pg_model = Mock()
    read = Mock()
    write = Mock()
    read.get_by_uuid.return_value = es_model
    write.get_by_uuid.return_value = pg_model
    es_model.invalidated.return_value = False
    pg_model.invalidated.return_value = False
    storage = PickStorage(read, write)
    with testConfig(request=dummy_request):
        model = storage.get_by_uuid('123')
    assert model is es_model
    assert dummy_request.datastore == 'elasticsearch'
    # Switches to PG when model invalidated.
    es_model.invalidated.return_value = True
    with testConfig(request=dummy_request):
        model = storage.get_by_uuid('123')
    assert model is pg_model
    assert dummy_request.datastore == 'database'


def test_pick_storage_get_by_unique_key_does_not_force_database_when_model_is_none(dummy_request):
    from pyramid.testing import testConfig
    from ..esstorage import PickStorage
    dummy_request.datastore = 'elasticsearch'
    es_model = Mock()
    pg_model = Mock()
    read = Mock()
    write = Mock()
    read.get_by_unique_key.return_value = es_model
    write.get_by_unique_key.return_value = pg_model
    es_model.invalidated.return_value = False
    pg_model.invalidated.return_value = False
    storage = PickStorage(read, write)
    with testConfig(request=dummy_request):
        model = storage.get_by_unique_key('accession', 'SNOFL000LSQ')
    assert model is es_model
    assert dummy_request.datastore == 'elasticsearch'
    # Set model to None.
    read.get_by_unique_key.return_value = None
    with testConfig(request=dummy_request):
        model = storage.get_by_unique_key('accession', 'SNOFL000LSQ')
    # Make sure it looks in PG and doesn't force request.
    assert model is pg_model
    assert dummy_request.datastore == 'elasticsearch'


def test_pick_storage_get_by_uuid_does_not_force_database_when_model_is_none(dummy_request):
    from pyramid.testing import testConfig
    from ..esstorage import PickStorage
    dummy_request.datastore = 'elasticsearch'
    es_model = Mock()
    pg_model = Mock()
    read = Mock()
    write = Mock()
    read.get_by_uuid.return_value = es_model
    write.get_by_uuid.return_value = pg_model
    es_model.invalidated.return_value = False
    pg_model.invalidated.return_value = False
    storage = PickStorage(read, write)
    with testConfig(request=dummy_request):
        model = storage.get_by_uuid('123')
    assert model is es_model
    assert dummy_request.datastore == 'elasticsearch'
    # Set model to None.
    read.get_by_uuid.return_value = None
    with testConfig(request=dummy_request):
        model = storage.get_by_uuid('123')
    # Make sure it looks in PG and doesn't force request.
    assert model is pg_model
    assert dummy_request.datastore == 'elasticsearch'



from pyramid.interfaces import PHASE2_CONFIG
from snovault import (
    TYPES,
    UPGRADER,
)
from snovault.upgrader import default_upgrade_finalizer

LATE = 10


def includeme(config):
    config.scan()

    def callback():
        """ add_upgrade for all item types
        """
        upgrader = config.registry[UPGRADER]
        types = config.registry[TYPES]
        for type_info in types.by_item_type.values():
            version = type_info.schema_version
            if version is not None:
                upgrader.add_upgrade(type_info.name, version)

    config.action('add_upgrades', callback, order=PHASE2_CONFIG)

    def default_upgrades():
        """ add_upgrade for all item types
        """
        upgrader = config.registry[UPGRADER]
        types = config.registry[TYPES]
        for type_info in types.by_item_type.values():
            if type_info.name not in upgrader:
                continue
            if not upgrader[type_info.name].upgrade_steps:
                upgrader[type_info.name].add_upgrade_step(run_finalizer)

    config.action('add_default_upgrades', default_upgrades, order=LATE)


@default_upgrade_finalizer
def finalizer(value, system, version):
    # Update the default properties
    value['schema_version'] = version


def run_finalizer(value, system):
    pass



from snovault import COLLECTIONS


def groupfinder(login, request):
    if '.' not in login:
        return None
    namespace, localname = login.split('.', 1)
    user = None

    collections = request.registry[COLLECTIONS]

    if namespace == 'remoteuser':
        if localname in ['EMBED', 'INDEXER']:
            return []
        elif localname in ['TEST', 'IMPORT', 'UPGRADE']:
            return ['group.admin']
        elif localname in ['TEST_SUBMITTER']:
            return ['group.submitter']
        elif localname in ['TEST_AUTHENTICATED']:
            return ['viewing_group.ENCODE']

    if namespace in ('mailto', 'remoteuser', 'webuser'):
        users = collections.by_item_type['user']
        try:
            user = users[localname]
        except KeyError:
            return None

    elif namespace == 'accesskey':
        access_keys = collections.by_item_type['access_key']
        try:
            access_key = access_keys[localname]
        except KeyError:
            return None

        if access_key.properties.get('status') in ('deleted', 'disabled'):
            return None

        userid = access_key.properties['user']
        user = collections.by_item_type['user'][userid]

    if user is None:
        return None

    user_properties = user.properties

    if user_properties.get('status') in ('deleted', 'disabled'):
        return None

    principals = ['userid.%s' % user.uuid]
    lab = user_properties.get('lab')
    if lab:
        principals.append('lab.%s' % lab)
    submits_for = user_properties.get('submits_for', [])
    principals.extend('lab.%s' % lab_uuid for lab_uuid in submits_for)
    principals.extend('submits_for.%s' % lab_uuid for lab_uuid in submits_for)
    if submits_for:
        principals.append('group.submitter')
    groups = user_properties.get('groups', [])
    principals.extend('group.%s' % group for group in groups)
    viewing_groups = user_properties.get('viewing_groups', [])
    principals.extend('viewing_group.%s' % group for group in viewing_groups)
    return principals



import pytest


def test_object_view_get_default(testapp, dummy_request, posted_targets_and_sources):
    res = testapp.get('/testing-link-targets/one/')
    assert 'reverse' in res.json


def test_object_view_get_skip_calcualted(testapp, dummy_request, posted_targets_and_sources):
    res = testapp.get('/testing-link-targets/one/?frame=object&skip_calculated=true')
    assert 'reverse' not in res.json


def test_object_view_get_object_with_select_calculated_properties_none(testapp, dummy_request, posted_targets_and_sources):
    res = testapp.get('/testing-link-targets/one/?frame=object_with_select_calculated_properties')
    assert 'reverse' not in res.json


def test_object_view_get_object_with_select_calculated_properties_reverse(testapp, dummy_request, posted_targets_and_sources):
    res = testapp.get('/testing-link-targets/one/?frame=object_with_select_calculated_properties&field=reverse')
    assert 'reverse' in res.json


def test_object_view_embed_default(dummy_request, threadlocals, posted_targets_and_sources):
    res = dummy_request.embed('/testing-link-targets/one/', '@@object')
    assert 'reverse' in res


def test_object_view_embed_skip_calculated(dummy_request, threadlocals, posted_targets_and_sources):
    res = dummy_request.embed(
        '/testing-link-targets/one/',
        '@@object?skip_calculated=true'
    )
    assert 'reverse' not in res


def test_object_view_embed_object_with_select_calculated_properties_none(dummy_request, threadlocals, posted_targets_and_sources):
    res = dummy_request.embed('/testing-link-targets/one/', '@@object_with_select_calculated_properties')
    assert 'reverse' not in res


def test_object_view_embed_object_with_select_calculated_properties_reverse(dummy_request, threadlocals, posted_targets_and_sources):
    res = dummy_request.embed(
        '/testing-link-targets/one/',
        '@@object_with_select_calculated_properties?field=reverse&field=nonsense'
    )
    assert 'reverse' in res


def test_calculated_should_render_property():
    from snovault.calculated import _should_render_property
    assert _should_render_property(None, None, 'abc')
    assert _should_render_property(None, [], 'abc')
    assert not _should_render_property([], None, 'abc')
    assert not _should_render_property([], [], 'abc')
    assert _should_render_property(['abc'], None, 'abc')
    assert _should_render_property(['abc', 'xyz'], None, 'abc')
    assert not _should_render_property(['xyz'], None, 'abc')
    assert not _should_render_property(['xyz'], None, 'abc')
    assert not _should_render_property(['xyz'], [], 'abc')
    assert not _should_render_property(['xyz'], ['abc'], 'abc')
    assert not _should_render_property(['xyz'], ['abc'], 'abc')
    assert _should_render_property(['abc', 'xyz'], [], 'abc')
    assert _should_render_property(['xyz'], [], 'xyz')


def test_filtered_object_view_get_default(testapp, dummy_request, posted_targets_and_sources):
    r = testapp.get('/testing-link-targets/one/')
    expected_keys = [
        'name',
        '@id',
        '@type',
        'uuid',
        'reverse',
        '@context'
    ]
    assert all(
        [
            e in r.json
            for e in expected_keys
        ]
    )


def test_filtered_object_view_get_filtered_object(testapp, dummy_request, posted_targets_and_sources):
    r = testapp.get('/testing-link-targets/one/@@filtered_object')
    expected_keys = [
        'name',
        '@id',
        '@type',
        'uuid',
        'reverse',
    ]
    assert all(
        [
            e in r.json
            for e in expected_keys
        ]
    )


def test_filtered_object_view_get_filtered_object_with_exclude(testapp, dummy_request, posted_targets_and_sources):
    r = testapp.get('/testing-link-targets/one/@@filtered_object?exclude=reverse')
    expected_keys = [
        'name',
        '@id',
        '@type',
        'uuid',
    ]
    assert all(
        [
            e in r.json
            for e in expected_keys
        ]
    )
    assert 'reverse' not in r.json


def test_filtered_object_view_get_filtered_object_with_more_exclude(testapp, dummy_request, posted_targets_and_sources):
    r = testapp.get('/testing-link-targets/one/@@filtered_object?exclude=reverse&exclude=@id')
    expected_keys = [
        'name',
        '@type',
        'uuid',
    ]
    assert all(
        [
            e in r.json
            for e in expected_keys
        ]
    )
    assert 'reverse' not in r.json
    assert '@id' not in r.json


def test_filtered_object_view_get_filtered_object_with_exclude_and_include(testapp, dummy_request, posted_targets_and_sources):
    r = testapp.get('/testing-link-targets/one/@@filtered_object?exclude=reverse&exclude=@id&include=uuid')
    expected_keys = [
        'uuid',
    ]
    assert all(
        [
            e in r.json
            for e in expected_keys
        ]
    )
    assert 'reverse' not in r.json
    assert '@id' not in r.json
    assert 'name' not in r.json
    assert '@type' not in r.json


def test_embedded_with_frame_custom_embed(testapp, dummy_request, posted_custom_embed_targets_and_sources):
    r = testapp.get('/testing-custom-embed-targets/one/')
    assert r.json['reverse'][0] == {
        'name': 'A',
        'status': 'current',
        'target': '/testing-custom-embed-targets/one/',
        '@id': '/testing-custom-embed-sources/16157204-8c8f-4672-a1a4-14f4b8021fcd/',
        '@type': ['TestingCustomEmbedSource', 'Item'],
        'uuid': '16157204-8c8f-4672-a1a4-14f4b8021fcd'
    }
    assert r.json['filtered_reverse'][0] == {
        'status': 'current',
        'uuid': '16157204-8c8f-4672-a1a4-14f4b8021fcd'
    }
    assert r.json['filtered_reverse1'][0] == {
        'name': 'A',
        'status': 'current',
        'target': '/testing-custom-embed-targets/one/',
        '@id': '/testing-custom-embed-sources/16157204-8c8f-4672-a1a4-14f4b8021fcd/'
    }
    assert r.json['reverse_uncalculated'][0] == {
        'name': 'A',
        'status': 'current',
        'target': '/testing-custom-embed-targets/one/'
    }



from pyramid.httpexceptions import HTTPNotFound
from pyramid.view import view_config
from .etag import etag_app_version_effective_principals
from .interfaces import (
    COLLECTIONS,
    TYPES,
)
from .util import mutated_schema


def includeme(config):
    config.add_route('schemas', '/profiles{slash:/?}')
    config.add_route('schema', '/profiles/{type_name}.json')
    config.add_route('schemap', '/profiles/{type_name}{slash:/?}')
    config.add_route('schemas_map', '/profiles-map/')
    config.add_route('schemas_titles', '/profiles-titles/')
    config.scan(__name__)


def _apply_permission(collection, request):
    def mutator(schema):
        if 'permission' in schema:
            if not request.has_permission(schema['permission'], collection):
                schema = schema.copy()
                schema['readonly'] = True
        return schema
    return mutator


def _annotated_schema(type_info, request):
    schema = type_info.schema.copy()
    schema['@type'] = ['JSONSchema']
    if type_info.factory is None:
        return schema

    collection = request.registry[COLLECTIONS][type_info.name]
    return mutated_schema(
        schema,
        _apply_permission(collection, request),
    )


@view_config(route_name='schema', request_method='GET',
             decorator=etag_app_version_effective_principals)
@view_config(route_name='schemap', request_method='GET',
             decorator=etag_app_version_effective_principals)
def schema(context, request):
    type_name = request.matchdict['type_name']
    types = request.registry[TYPES]
    try:
        type_info = types[type_name]
    except KeyError:
        raise HTTPNotFound(type_name)

    return _annotated_schema(type_info, request)


@view_config(route_name='schemas', request_method='GET',
             decorator=etag_app_version_effective_principals)
def schemas(context, request):
    types = request.registry[TYPES]
    schemas = {}
    for type_info in types.by_item_type.values():
        name = type_info.name
        schemas[name] = _annotated_schema(type_info, request)

    schemas['_subtypes'] = subtypes = {}
    schemas['@type'] = ['JSONSchemas']
    for name, type_info in types.abstract.items():
        subtypes[name] = type_info.subtypes

    return schemas


@view_config(route_name='schemas_map', request_method='GET',
             decorator=etag_app_version_effective_principals)
def schemas_map(context, request):
    types = request.registry[TYPES]
    profiles_map = {}
    for type_info in types.by_item_type.values():
        if 'id' in type_info.schema:
            profiles_map[type_info.name] = type_info.schema['id']
    profiles_map['@type'] = ['JSONSchemas']
    return profiles_map


@view_config(route_name='schemas_titles', request_method='GET',
             decorator=etag_app_version_effective_principals)
def schemas_titles(context, request):  # pylint: disable=unused-argument
    '''Return mapping of all schema @types and their corresponding titles'''
    types = request.registry[TYPES]
    profiles_titles = {
        type_info.name: type_info.schema['title']
        for type_info in types.by_item_type.values()
        if 'title' in type_info.schema
    }
    profiles_titles['@type'] = ['JSONSchemas']
    return profiles_titles



from snovault import upgrade_step

''' Note these are not relevant to anything beyond testing upgrader '''


@upgrade_step('snowball', '', '2')
@upgrade_step('snowfort', '', '2')
def snowset_0_2(value, system):
    # example upgrade for tests
    if 'status' in value:
        if value['status'] == 'DELETED':
            value['status'] = 'deleted'
        elif value['status'] == 'CURRENT':
            value['status'] = 'submitted'  # there is a dependency on date_released+"released"



from pyramid.security import (
    Authenticated,
    Everyone,
    principals_allowed_by_permission,
)
from pyramid.traversal import resource_path
from pyramid.view import view_config
from .resources import Item


def includeme(config):
    config.scan(__name__)


@view_config(context=Item, name='index-data', permission='index', request_method='GET')
def item_index_data(context, request):
    uuid = str(context.uuid)
    properties = context.upgrade_properties()
    links = context.links(properties)
    unique_keys = context.unique_keys(properties)

    principals_allowed = {}
    for permission in ('view', 'edit', 'audit'):
        principals = principals_allowed_by_permission(context, permission)
        if principals is Everyone:
            principals = [Everyone]
        elif Everyone in principals:
            principals = [Everyone]
        elif Authenticated in principals:
            principals = [Authenticated]
        # Filter our roles
        principals_allowed[permission] = [
            p for p in sorted(principals) if not p.startswith('role.')
        ]

    path = resource_path(context)
    paths = {path}
    collection = context.collection

    if collection.unique_key in unique_keys:
        paths.update(
            resource_path(collection, key)
            for key in unique_keys[collection.unique_key])

    for base in (collection, request.root):
        for key_name in ('accession', 'alias'):
            if key_name not in unique_keys:
                continue
            paths.add(resource_path(base, uuid))
            paths.update(
                resource_path(base, key)
                for key in unique_keys[key_name])

    path = path + '/'
    embedded = request.embed(path, '@@embedded')
    object = request.embed(path, '@@object')

    audit = request.embed(path, '@@audit')['audit']

    document = {
        'audit': audit,
        'embedded': embedded,
        'embedded_uuids': sorted(request._embedded_uuids),
        'item_type': context.type_info.item_type,
        'linked_uuids': sorted(request._linked_uuids),
        'links': links,
        'object': object,
        'paths': sorted(paths),
        'principals_allowed': principals_allowed,
        'properties': properties,
        'propsheets': {
            name: context.propsheets[name]
            for name in context.propsheets.keys() if name != ''
        },
        'tid': context.tid,
        'unique_keys': unique_keys,
        'uuid': uuid,
    }

    return document



import pytest


@pytest.fixture()
def dummy_parent(dummy_request):
    from pyramid.testing import DummyResource
    from pyramid.security import Allow
    from snosearch.parsers import ParamsParser
    from snosearch.queries import AbstractQueryFactory
    from snovault.elasticsearch.interfaces import ELASTIC_SEARCH
    from elasticsearch import Elasticsearch
    dummy_request.registry[ELASTIC_SEARCH] = Elasticsearch()
    dummy_request.context = DummyResource()
    dummy_request.context.__acl__ = lambda: [(Allow, 'group.submitter', 'search_audit')]
    class DummyParent():
        def __init__(self):
            self._meta = {}
            self.response = {}
    dp = DummyParent()
    pp = ParamsParser(dummy_request)
    dp._meta = {
        'params_parser': pp,
        'query_builder': AbstractQueryFactory(pp)
    }
    return dp


def test_searches_fields_non_sortable_response_field(dummy_parent):
    dummy_parent._meta['params_parser']._request.environ['QUERY_STRING'] = (
        'type=TestingSearchSchema'
    )
    from snovault.elasticsearch.searches.fields import NonSortableResponseField
    nrf = NonSortableResponseField()
    r = nrf.render(parent=dummy_parent)
    assert r['non_sortable'] == ['pipeline_error_detail', 'description', 'notes']



import pytest


def raising_checker(value, system):
    from snovault.auditor import AuditFailure
    if not value.get('checker1'):
        raise AuditFailure('testchecker', 'Missing checker1')


def returning_checker(value, system):
    from snovault.auditor import AuditFailure
    if not value.get('checker1'):
        return AuditFailure('testchecker', 'Missing checker1')


def yielding_checker(value, system):
    from snovault.auditor import AuditFailure
    if not value.get('checker1'):
        yield AuditFailure('testchecker', 'Missing checker1')


def has_condition1(value, system):
    return value.get('condition1')


@pytest.fixture(autouse=True)
def autouse_external_tx(external_tx):
    pass


@pytest.fixture(params=[
    raising_checker,
    returning_checker,
    yielding_checker,
])
def auditor(request):
    from snovault.auditor import Auditor
    auditor = Auditor()
    auditor.add_audit_checker(request.param, 'test')
    return auditor


@pytest.fixture
def auditor_conditions():
    from snovault.auditor import Auditor
    auditor = Auditor()
    auditor.add_audit_checker(raising_checker, 'test', has_condition1)
    return auditor


@pytest.fixture
def dummy_request(registry):
    from pyramid.testing import DummyRequest
    _embed = {}
    request = DummyRequest(registry=registry, _embed=_embed, embed=lambda path: _embed[path])
    return request


def test_audit_pass(auditor, dummy_request):
    value = {'checker1': True}
    dummy_request._embed['/foo/@@embedded'] = value
    errors = auditor.audit(request=dummy_request, path='/foo/', types='test')
    assert errors == []


def test_audit_failure(auditor, dummy_request):
    value = {}
    dummy_request._embed['/foo/@@embedded'] = value
    error, = auditor.audit(request=dummy_request, path='/foo/', types='test')
    assert error['detail'] == 'Missing checker1'
    assert error['category'] == 'testchecker'
    assert error['level'] == 0
    assert error['path'] == '/foo/'


def test_audit_conditions(auditor_conditions, dummy_request):
    value = {}
    dummy_request._embed['/foo/@@embedded'] = value
    assert auditor_conditions.audit(request=dummy_request, path='/foo/', types='test') == []
    value = {'condition1': True}
    dummy_request._embed['/foo/@@embedded'] = value
    error, = auditor_conditions.audit(request=dummy_request, path='/foo/', types='test')
    assert error['detail'] == 'Missing checker1'
    assert error['category'] == 'testchecker'
    assert error['level'] == 0
    assert error['path'] == '/foo/'


def test_declarative_config(dummy_request):
    from snovault.interfaces import AUDITOR
    from pyramid.config import Configurator
    config = Configurator()
    config.include('snovault.config')
    config.include('snovault.auditor')
    config.include('snovault.elasticsearch.searches.configs')
    config.include('.testing_auditor')
    config.commit()

    auditor = config.registry[AUDITOR]
    value = {'condition1': True}
    dummy_request._embed['/foo/@@embedded'] = value
    error, = auditor.audit(request=dummy_request, path='/foo/', types='TestingLinkSource')
    assert error['detail'] == 'Missing checker1'
    assert error['category'] == 'testchecker'
    assert error['level'] == 0
    assert error['path'] == '/foo/'


def test_link_target_audit_fail(testapp):
    target = {'uuid': '775795d3-4410-4114-836b-8eeecf1d0c2f', 'status': 'CHECK'}
    testapp.post_json('/testing_link_target', target, status=201)
    res = testapp.get('/%s/@@index-data' % target['uuid']).maybe_follow()
    errors_dict = res.json['audit']
    errors_list = []
    for error_type in errors_dict:
        errors_list.extend(errors_dict[error_type])
    errors = [e for e in errors_list if e['name'] == 'testing_link_target_status']
    error, = errors
    assert error['detail'] == 'Missing reverse items'
    assert error['category'] == 'status'
    assert error['level'] == 0
    assert error['path'] == res.json['object']['@id']


def test_link_target_audit_pass(testapp):
    target = {'uuid': '775795d3-4410-4114-836b-8eeecf1d0c2f', 'status': 'CHECK'}
    testapp.post_json('/testing_link_target', target, status=201)
    source = {'uuid': '16157204-8c8f-4672-a1a4-14f4b8021fcd', 'target': target['uuid']}
    testapp.post_json('/testing_link_source', source, status=201)
    res = testapp.get('/%s/@@index-data' % target['uuid']).maybe_follow()
    errors_dict = res.json['audit']
    errors_list = []
    for error_type in errors_dict:
        errors_list.extend(errors_dict[error_type])
    errors = [e for e in errors_list if e['name'] == 'testing_link_target_status']
    assert errors == []


def test_audit_inherit_error(testapp):
    target = {'uuid': '775795d3-4410-4114-836b-8eeecf1d0c2f', 'status': 'CHECK'}
    testapp.post_json('/testing_link_target', target, status=201)
    source = {'uuid': '16157204-8c8f-4672-a1a4-14f4b8021fcd', 'target': target['uuid'], 'condition1': True}
    # condition1 - auto fails the audit on source
    testapp.post_json('/testing_link_source', source, status=201)

    res = testapp.get('/%s/@@index-data' % target['uuid']).maybe_follow()
    # index TARGET
    errors_dict = res.json['audit']
    errors_list = []
    for error_type in errors_dict:
        errors_list.extend(errors_dict[error_type])
    errors = [e for e in errors_list if e['name'] == 'checker1']
    assert len(errors) == 1
    error = errors[0]
    assert error['detail'] == 'Missing checker1'
    assert error['category'] == 'testchecker'
    assert error['level'] == 0
    assert error['path'] == '/testing-link-sources/16157204-8c8f-4672-a1a4-14f4b8021fcd/'


def test_audit_inherit_no_error(testapp):
    target = {'uuid': '775795d3-4410-4114-836b-8eeecf1d0c2f', 'status': 'CHECK'}
    testapp.post_json('/testing_link_target', target, status=201)
    source = {'uuid': '16157204-8c8f-4672-a1a4-14f4b8021fcd', 'target': target['uuid'], 'condition1': False}
    testapp.post_json('/testing_link_source', source, status=201)

    res = testapp.get('/%s/@@index-data' % target['uuid']).maybe_follow()
    errors_dict = res.json['audit']
    for error_type in errors_dict:
        assert not errors_dict[error_type]['name'] == 'checker1'



from pytest import fixture


COLLECTION_URL = '/testing-server-defaults'


@fixture
def extra_environ(admin):
    email = admin['email']
    return {'REMOTE_USER': str(email)}


@fixture
def default_content(testapp, external_tx):
    res = testapp.post_json(COLLECTION_URL, {}, status=201)
    return {'@id': res.json['@graph'][0]['@id']}


def test_server_defaults(admin, anontestapp, extra_environ):
    res = anontestapp.post_json(
        COLLECTION_URL, {}, status=201,
        extra_environ=extra_environ,
    )
    item = res.json['@graph'][0]
    assert item['now'].startswith('2')
    assert item['user'] == admin['@id']
    assert item['accession'].startswith('SNO')

    anontestapp.patch_json(
        res.location, {}, status=200,
        extra_environ=extra_environ,
    )


@fixture(scope='session')
def test_accession_app(request, check_constraints, zsa_savepoints, app_settings):
    from snowflakes import main
    app_settings = app_settings.copy()
    app_settings['accession_factory'] = 'snowflakes.server_defaults.test_accession'
    return main({}, **app_settings)


@fixture
def test_accession_anontestapp(request, test_accession_app, external_tx, zsa_savepoints):
    '''TestApp with JSON accept header.
    '''
    from webtest import TestApp
    environ = {
        'HTTP_ACCEPT': 'application/json',
    }
    return TestApp(test_accession_app, environ)


@fixture
def test_accession_authtestapp(request, test_accession_app, external_tx, zsa_savepoints):
    '''TestApp with JSON accept header.
    '''
    from webtest import TestApp
    environ = {
        'HTTP_ACCEPT': 'application/json',
        'REMOTE_USER': 'TEST_AUTHENTICATED',
    }
    return TestApp(test_accession_app, environ)


def test_test_accession_server_defaults(admin, test_accession_anontestapp, extra_environ):
    res = test_accession_anontestapp.post_json(
        COLLECTION_URL, {}, status=201,
        extra_environ=extra_environ,
    )
    item = res.json['@graph'][0]
    assert item['accession'].startswith('TSTSS')

    test_accession_anontestapp.patch_json(
        res.location, {}, status=200,
        extra_environ=extra_environ,
    )



import pytest


from snovault.util import _get_calculated_properties_from_paths
from snovault.util import select_distinct_values


# Test item with calculated property.
COLLECTION_URL = '/testing-link-targets/'


def test_get_calculated_properties_from_paths(dummy_request):
    paths = [
        COLLECTION_URL + 'ABC123/'
    ]
    calculated_properties = _get_calculated_properties_from_paths(dummy_request, paths)
    assert 'reverse' in calculated_properties


def test_not_a_collection_get_calculated_properties_from_paths(dummy_request):
    paths = [
        '/yxa/ABC123/'
    ]
    calculated_properties = _get_calculated_properties_from_paths(dummy_request, paths)
    assert not calculated_properties


def test_unformed_path_get_calculated_properties_from_paths(dummy_request):
    paths = [
        'testing-link-targets/ABC123/'
    ]
    calculated_properties = _get_calculated_properties_from_paths(dummy_request, paths)
    assert not calculated_properties


def test_select_distinct_values_returns_calculated(dummy_request, threadlocals, posted_targets_and_sources):
    distinct_values = select_distinct_values(dummy_request, 'reverse', *['/testing-link-targets/one/'])
    assert '/testing-link-sources/16157204-8c8f-4672-a1a4-14f4b8021fcd/' in distinct_values


def test_select_distinct_values_uses_calculated(dummy_request, threadlocals, posted_targets_and_sources, mocker):
    mocker.patch.object(dummy_request, 'embed')
    select_distinct_values(dummy_request, 'reverse', *['/testing-link-targets/one/'])
    dummy_request.embed.assert_called_with('/testing-link-targets/one/', '@@object')


def test_select_distinct_values_skips_calculated(dummy_request, threadlocals, posted_targets_and_sources, mocker):
    mocker.patch.object(dummy_request, 'embed')
    select_distinct_values(dummy_request, 'name', *['/testing-link-targets/one/'])
    dummy_request.embed.assert_called_with('/testing-link-targets/one/', '@@object?skip_calculated=true')


def test_types_utils_ensure_list():
    from snovault.util import ensure_list_and_filter_none
    assert ensure_list_and_filter_none('abc') == ['abc']
    assert ensure_list_and_filter_none(['abc']) == ['abc']
    assert ensure_list_and_filter_none({'a': 'b'}) == [{'a': 'b'}]
    assert ensure_list_and_filter_none([{'a': 'b'}, {'c': 'd'}]) == [{'a': 'b'}, {'c': 'd'}]
    assert ensure_list_and_filter_none([{'a': 'b'}, {'c': 'd'}, None]) == [{'a': 'b'}, {'c': 'd'}]


def test_types_utils_take_one_or_return_none():
    from snovault.util import take_one_or_return_none
    assert take_one_or_return_none(['just one']) == 'just one'
    assert take_one_or_return_none(['one', 'and', 'two']) is None
    assert take_one_or_return_none('just one') is None


def test_util_path_init():
    from snovault.util import Path
    p = Path('abc')
    assert isinstance(p, Path)
    assert p.path == 'abc'
    assert p._frame == '@@object'
    assert p._params == {
        'include': [],
        'exclude': []
    }
    assert p.frame == '@@object'


def test_util_path_include_and_exclude_and_frame():
    from snovault.util import Path
    p = Path('/snoflakes/SNS123ABC/', include=['@id', '@type', 'uuid'])
    assert p.frame == '@@filtered_object?include=%40id&include=%40type&include=uuid'
    assert p.path == '/snoflakes/SNS123ABC/'
    p = Path('/snoflakes/SNS123ABC/', exclude=['description'])
    assert p.frame == '@@filtered_object?exclude=description'
    p = Path('/snoflakes/SNS123ABC/', include=['title', 'name'], exclude=['description', 'notes'])
    assert p.frame == '@@filtered_object?include=title&include=name&exclude=description&exclude=notes'
    p = Path(
        '/snoflakes/SNS123ABC/',
        frame='@@embedded',
        include=['title', 'name'],
        exclude=['description', 'notes']
    )
    assert p.frame == '@@filtered_object?include=title&include=name&exclude=description&exclude=notes'
    p = Path(
        '/snoflakes/SNS123ABC/',
        frame='@@embedded',
    )
    assert p.frame == '@@embedded'

def test_util_path_build_frame():
    from snovault.util import Path
    p = Path('/snoflakes/SNS123ABC/', include=['@id', '@type', 'uuid'])
    p._params = {'field': ['a', 'b', 'c']}
    assert p._build_frame() == '@@filtered_object?field=a&field=b&field=c'


def test_util_path_expand_path_with_frame(dummy_request, threadlocals, posted_custom_embed_targets_and_sources):
    from snovault.util import Path
    properties = {'targets': ['/testing-custom-embed-targets/one/']}
    p = Path('targets')
    p.expand_path_with_frame(dummy_request, properties, p.path, p.frame)
    assert properties['targets'][0] == {
        'name': 'one',
        '@id': '/testing-custom-embed-targets/one/',
        '@type': ['TestingCustomEmbedTarget', 'Item'],
        'uuid': '775795d3-4410-4114-836b-8eeecf1d0c2f',
        'reverse': ['/testing-custom-embed-sources/16157204-8c8f-4672-a1a4-14f4b8021fcd/'],
        'filtered_reverse': ['/testing-custom-embed-sources/16157204-8c8f-4672-a1a4-14f4b8021fcd/'],
        'filtered_reverse1': ['/testing-custom-embed-sources/16157204-8c8f-4672-a1a4-14f4b8021fcd/'],
        'reverse_uncalculated': ['/testing-custom-embed-sources/16157204-8c8f-4672-a1a4-14f4b8021fcd/']
    }
    p = Path('targets.reverse')
    p.expand_path_with_frame(dummy_request, properties, p.path, p.frame)
    assert properties['targets'][0]['reverse'][0] == {
        'name': 'A',
        'status': 'current',
        'target': '/testing-custom-embed-targets/one/',
        '@id': '/testing-custom-embed-sources/16157204-8c8f-4672-a1a4-14f4b8021fcd/',
        '@type': ['TestingCustomEmbedSource', 'Item'], 'uuid': '16157204-8c8f-4672-a1a4-14f4b8021fcd'
    }
    p = Path('targets.reverse.target', include=['name', 'uuid'])
    p.expand_path_with_frame(dummy_request, properties, p.path, p.frame)
    assert properties['targets'][0]['reverse'][0]['target'] == {
        'name': 'one',
        'uuid': '775795d3-4410-4114-836b-8eeecf1d0c2f'
    }


def test_util_path_expand_path_with_frame_include(dummy_request, threadlocals, posted_custom_embed_targets_and_sources):
    from snovault.util import Path
    properties = {'targets': ['/testing-custom-embed-targets/one/']}
    p = Path('targets.reverse.target')
    p.expand_path_with_frame(dummy_request, properties, p.path, p.frame)
    assert properties['targets'][0]['reverse'][0]['target'] == {
        'name': 'one',
        '@id': '/testing-custom-embed-targets/one/',
        '@type': ['TestingCustomEmbedTarget', 'Item'],
        'uuid': '775795d3-4410-4114-836b-8eeecf1d0c2f',
        'reverse': ['/testing-custom-embed-sources/16157204-8c8f-4672-a1a4-14f4b8021fcd/'],
        'filtered_reverse': ['/testing-custom-embed-sources/16157204-8c8f-4672-a1a4-14f4b8021fcd/'],
        'filtered_reverse1': ['/testing-custom-embed-sources/16157204-8c8f-4672-a1a4-14f4b8021fcd/'],
        'reverse_uncalculated': ['/testing-custom-embed-sources/16157204-8c8f-4672-a1a4-14f4b8021fcd/']
    }
    properties = {'targets': ['/testing-custom-embed-targets/one/']}
    p = Path('targets.reverse.target', include=['@id'])
    p.expand_path_with_frame(dummy_request, properties, p.path, p.frame)
    assert properties == {
        'targets': [
            {'@id': '/testing-custom-embed-targets/one/'}
        ]
    }
    properties = {'targets': ['/testing-custom-embed-targets/one/']}
    p = Path('targets.reverse.target', include=['@id', 'reverse'])
    p.expand_path_with_frame(dummy_request, properties, p.path, p.frame)
    assert properties == {
        'targets': [
            {
                '@id': '/testing-custom-embed-targets/one/',
                'reverse': [
                    {
                        '@id': '/testing-custom-embed-sources/16157204-8c8f-4672-a1a4-14f4b8021fcd/'
                    }
                ]
            }
        ]
    }
    properties = {'targets': ['/testing-custom-embed-targets/one/']}
    p = Path('targets.reverse.target', include=['@id', 'reverse', 'target'])
    p.expand_path_with_frame(dummy_request, properties, p.path, p.frame)
    assert properties == {
        'targets': [
            {
                '@id': '/testing-custom-embed-targets/one/',
                'reverse': [
                    {
                        'target': {
                            '@id': '/testing-custom-embed-targets/one/',
                            'reverse': [
                                '/testing-custom-embed-sources/16157204-8c8f-4672-a1a4-14f4b8021fcd/'
                            ]
                        },
                        '@id': '/testing-custom-embed-sources/16157204-8c8f-4672-a1a4-14f4b8021fcd/'
                    }
                ]
            }
        ]
    }


def test_util_path_expand_path_with_frame_slim_embedding(dummy_request, threadlocals, posted_custom_embed_targets_and_sources):
    from snovault.util import Path
    properties = {'targets': ['/testing-custom-embed-targets/one/']}
    p = Path('targets', include=['reverse'])
    p.expand_path_with_frame(dummy_request, properties, p.path, p.frame)
    p = Path('targets.reverse', include=['target'])
    p.expand_path_with_frame(dummy_request, properties, p.path, p.frame)
    p = Path('targets.reverse.target', include=['name'])
    p.expand_path_with_frame(dummy_request, properties, p.path, p.frame)
    assert properties == {
        'targets': [
            {
                'reverse': [
                    {
                        'target': {
                            'name': 'one'
                        }
                    }
                ]
            }
        ]
    }


def test_util_path_expand_explicit_frame(dummy_request, threadlocals, posted_custom_embed_targets_and_sources):
    from snovault.util import Path
    properties = {'targets': ['/testing-custom-embed-targets/one/']}
    p = Path('targets', frame='@@embedded')
    p.expand(dummy_request, properties)
    assert properties == {
        'targets': [
            {
                'name': 'one',
                '@id': '/testing-custom-embed-targets/one/',
                '@type': ['TestingCustomEmbedTarget', 'Item'],
                'uuid': '775795d3-4410-4114-836b-8eeecf1d0c2f',
                'reverse': [
                    {
                        'name': 'A',
                        'status': 'current',
                        'target': '/testing-custom-embed-targets/one/',
                        '@id': '/testing-custom-embed-sources/16157204-8c8f-4672-a1a4-14f4b8021fcd/',
                        '@type': ['TestingCustomEmbedSource', 'Item'],
                        'uuid': '16157204-8c8f-4672-a1a4-14f4b8021fcd'}
                ],
                'filtered_reverse': [
                    {
                        'status': 'current',
                        'uuid': '16157204-8c8f-4672-a1a4-14f4b8021fcd'
                    }
                ],
                'filtered_reverse1': [
                    {
                        'name': 'A',
                        'status': 'current',
                        'target': '/testing-custom-embed-targets/one/',
                        '@id': '/testing-custom-embed-sources/16157204-8c8f-4672-a1a4-14f4b8021fcd/'
                    }
                ],
                'reverse_uncalculated': [
                    {
                        'name': 'A',
                        'status': 'current',
                        'target': '/testing-custom-embed-targets/one/'
                    }
                ]
            }
        ]
    }


def test_util_path_expand_exclude(dummy_request, threadlocals, posted_custom_embed_targets_and_sources):
    from snovault.util import Path
    properties = {'targets': ['/testing-custom-embed-targets/one/']}
    p = Path('targets', exclude=['reverse', 'filtered_reverse', 'reverse_uncalculated', '@type'])
    p.expand(dummy_request, properties)
    assert properties == {
        'targets': [
            {
                'name': 'one',
                '@id': '/testing-custom-embed-targets/one/',
                'uuid': '775795d3-4410-4114-836b-8eeecf1d0c2f',
                'filtered_reverse1': [
                    '/testing-custom-embed-sources/16157204-8c8f-4672-a1a4-14f4b8021fcd/'
                ]
            }
        ]
    }


def test_get_uuids_from_file():
    from snovault.util import get_uuids_from_file
    from pkg_resources import resource_filename
    small_db_path = resource_filename('snowflakes', 'tests/data/inserts/small_db.tsv')
    assert get_uuids_from_file(small_db_path) == [
        '7a5e9183-b52f-4f75-9708-8e077b086b4e',
        'e2f35c88-a792-4dea-b5d2-30dc52ed2495'
    ]



import binascii

from os import urandom
from datetime import datetime
from pytz import timezone

from redis import StrictRedis


def base_result(local_store):
    local_dt = datetime.now(timezone(local_store.local_tz))
    return {
        '@type': ['result'],
        'utc_now': str(datetime.utcnow()),
        'lcl_now': f"{local_store.local_tz}: {local_dt}",
    }


class LocalStoreClient():
    '''
    Light redis wrapper and redis examples
    - get_tag function was added to return hex str
    - Can access client directly for full functionality
    '''
    def __init__(self, **kwargs):
        self.local_tz = kwargs.get('local_tz', 'GMT')
        self.client = StrictRedis(
            encoding='utf-8',
            decode_responses=True,
            db=kwargs['db_index'],
            host=kwargs['host'],
            port=kwargs['port'],
            socket_timeout=kwargs['socket_timeout'],
        )

    @staticmethod
    def get_tag(tag, num_bytes=2):
        '''
        Tags are the tag plus a random hex bytes string
        - Bytes string length is 2 * num bytes
        '''
        rand_hex_str = binascii.b2a_hex(urandom(num_bytes)).decode('utf-8')
        return f"{tag}:{rand_hex_str}"

    def ping(self):
        return self.client.ping()

    def dict_get(self, key):
        return self.client.hgetall(key)

    def dict_set(self, key, hash_dict):
        for k, v in hash_dict.items():
            self.client.hset(name=key, key=k, value=v)
    
    def get_tag_keys(self, tag):
        return self.client.keys(f"{tag}:*")

    def item_get(self, key):
        return self.client.get(key)
    
    def item_set(self, key, item):
        return self.client.set(key, item)

    def list_add(self, key, item):
        return self.client.lpush(key, item)
    
    def list_get(self, key, start=0, stop=-1):
        return self.client.lrange(key, start, stop)



import pytest


def _type_length():
    # Not a fixture as we need to parameterize tests on this
    from ..loadxl import ORDER
    from pkg_resources import resource_stream
    import codecs
    import json
    utf8 = codecs.getreader("utf-8")
    return {
        name: len(json.load(utf8(resource_stream('snowflakes', 'tests/data/inserts/%s.json' % name))))
        for name in ORDER
    }


TYPE_LENGTH = _type_length()

PUBLIC_COLLECTIONS = [
    'lab',
    'award',
]


def test_home(anonhtmltestapp):
    res = anonhtmltestapp.get('/', status=200)
    assert res.body.startswith(b'<!DOCTYPE html>')


def test_home_json(testapp):
    res = testapp.get('/', status=200)
    assert res.json['@type']


def test_home_app_version(testapp):
    res = testapp.get('/', status=200)
    assert 'app_version' in res.json


def test_vary_html(anonhtmltestapp):
    res = anonhtmltestapp.get('/', status=200)
    assert res.vary is not None
    assert 'Accept' in res.vary


def test_vary_json(anontestapp):
    res = anontestapp.get('/', status=200)
    assert res.vary is not None
    assert 'Accept' in res.vary


@pytest.mark.parametrize('item_type', [k for k in TYPE_LENGTH if k != 'user'])
def test_collections_anon(workbook, anontestapp, item_type):
    res = anontestapp.get('/' + item_type).follow(status=200)
    assert '@graph' in res.json


@pytest.mark.parametrize('item_type', [k for k in TYPE_LENGTH if k != 'user'])
def test_html_collections_anon(workbook, anonhtmltestapp, item_type):
    res = anonhtmltestapp.get('/' + item_type).follow(status=200)
    assert res.body.startswith(b'<!DOCTYPE html>')


@pytest.mark.parametrize('item_type', TYPE_LENGTH)
def test_html_collections(workbook, htmltestapp, item_type):
    res = htmltestapp.get('/' + item_type).follow(status=200)
    assert res.body.startswith(b'<!DOCTYPE html>')


@pytest.mark.slow
@pytest.mark.parametrize('item_type', TYPE_LENGTH)
def test_html_pages(workbook, testapp, htmltestapp, item_type):
    res = testapp.get('/%s?limit=all' % item_type).follow(status=200)
    for item in res.json['@graph']:
        res = htmltestapp.get(item['@id'])
        assert res.body.startswith(b'<!DOCTYPE html>')


@pytest.mark.slow
@pytest.mark.parametrize('item_type', [k for k in TYPE_LENGTH if k != 'user'])
def test_html_server_pages(workbook, item_type, wsgi_server):
    from webtest import TestApp
    testapp = TestApp(wsgi_server)
    res = testapp.get(
        '/%s?limit=all' % item_type,
        headers={'Accept': 'application/json'},
    ).follow(
        status=200,
        headers={'Accept': 'application/json'},
    )
    for item in res.json['@graph']:
        res = testapp.get(item['@id'], status=200)
        assert res.body.startswith(b'<!DOCTYPE html>')
        assert b'Internal Server Error' not in res.body


@pytest.mark.parametrize('item_type', TYPE_LENGTH)
def test_json(testapp, item_type):
    res = testapp.get('/' + item_type).follow(status=200)
    assert res.json['@type']


def test_json_basic_auth(anonhtmltestapp):
    from base64 import b64encode
    from pyramid.compat import ascii_native_
    url = '/'
    value = "Authorization: Basic %s" % ascii_native_(b64encode(b'nobody:pass'))
    res = anonhtmltestapp.get(url, headers={'Authorization': value}, status=401)
    assert res.content_type == 'application/json'


def test_load_sample_data(
        award,
        lab,
        submitter,
        snowball,
        snowflake,
        ):
    assert True, 'Fixtures have loaded sample data'


def test_abstract_collection(testapp, snowball):
    testapp.get('/Snowset/{accession}'.format(**snowball))
    testapp.get('/snowsets/{accession}'.format(**snowball))


def test_views_object_with_select_calculated_properties(workbook, testapp):
    r = testapp.get('/snowballs/')
    at_id = r.json['@graph'][0]['@id']
    r = testapp.get(at_id + '?frame=object')
    assert 'test_calculated' in r.json
    assert 'another_test_calculated' in r.json
    assert 'conditional_test_calculated' in r.json
    r = testapp.get(at_id + '?frame=object&skip_calculated=true')
    assert 'test_calculated' not in r.json
    assert 'another_test_calculated' not in r.json
    assert 'conditional_test_calculated' not in r.json
    r = testapp.get(at_id + '?frame=object_with_select_calculated_properties')
    assert 'test_calculated' not in r.json
    assert 'another_test_calculated' not in r.json
    assert 'conditional_test_calculated' not in r.json
    r = testapp.get(at_id + '?frame=object_with_select_calculated_properties&field=test_calculated')
    assert 'test_calculated' in r.json
    assert 'another_test_calculated' not in r.json
    assert 'conditional_test_calculated' not in r.json
    r = testapp.get(
        at_id + '?frame=object_with_select_calculated_properties&field=test_calculated&field=another_test_calculated'
    )
    assert 'test_calculated' in r.json
    assert 'another_test_calculated' in r.json
    assert 'conditional_test_calculated' not in r.json
    r = testapp.get(
        at_id + '?frame=object_with_select_calculated_properties&field=test_calculated&field=conditional_test_calculated'
    )
    assert 'test_calculated' in r.json
    assert 'another_test_calculated' not in r.json
    assert 'conditional_test_calculated' in r.json


@pytest.mark.slow
@pytest.mark.parametrize(('item_type', 'length'), TYPE_LENGTH.items())
def test_load_workbook(workbook, testapp, item_type, length):
    # testdata must come before testapp in the funcargs list for their
    # savepoints to be correctly ordered.
    res = testapp.get('/%s/?limit=all' % item_type).maybe_follow(status=200)
    assert len(res.json['@graph']) == length


@pytest.mark.slow
def test_collection_limit(workbook, testapp):
    res = testapp.get('/awards/?limit=2', status=200)
    assert len(res.json['@graph']) == 2


def test_collection_post(testapp):
    item = {
        'name': 'NIS39393',
        'title': 'Grant to make snow',
        'project': 'ENCODE',
        'rfa': 'ENCODE3',
    }
    return testapp.post_json('/award', item, status=201)


def test_collection_post_bad_json(testapp):
    item = {'foo': 'bar'}
    res = testapp.post_json('/award', item, status=422)
    assert res.json['errors']


def test_collection_post_malformed_json(testapp):
    item = '{'
    headers = {'Content-Type': 'application/json'}
    res = testapp.post('/award', item, status=400, headers=headers)
    assert res.json['detail'].startswith('Expecting')


def test_collection_post_missing_content_type(testapp):
    item = '{}'
    testapp.post('/award', item, status=415)


def test_collection_post_bad_(anontestapp):
    from base64 import b64encode
    from pyramid.compat import ascii_native_
    value = "Authorization: Basic %s" % ascii_native_(b64encode(b'nobody:pass'))
    anontestapp.post_json('/award', {}, headers={'Authorization': value}, status=401)


def test_collection_actions_filtered_by_permission(workbook, testapp, anontestapp):
    res = testapp.get('/pages/')
    assert any(action for action in res.json.get('actions', []) if action['name'] == 'add')

    res = anontestapp.get('/pages/')
    assert not any(action for action in res.json.get('actions', []) if action['name'] == 'add')


def test_item_actions_filtered_by_permission(testapp, authenticated_testapp, award):
    location = award['@id']

    res = testapp.get(location)
    assert any(action for action in res.json.get('actions', []) if action['name'] == 'edit')

    res = authenticated_testapp.get(location)
    assert not any(action for action in res.json.get('actions', []) if action['name'] == 'edit')


def test_collection_put(testapp, execute_counter):
    initial = {
        'name': 'NIS39393',
        'title': 'Grant to make snow',
        'project': 'ENCODE',
        'rfa': 'ENCODE3',
    }
    item_url = testapp.post_json('/award', initial).location

    with execute_counter.expect(1):
        item = testapp.get(item_url).json

    for key in initial:
        assert item[key] == initial[key]

    update = {
        'name': 'NIS440404',
        'title': 'Frozen wastland',
        'project': 'ENCODE',
        'rfa': 'ENCODE3',
    }
    testapp.put_json(item_url, update, status=200)

    res = testapp.get('/' + item['uuid']).follow().json

    for key in update:
        assert res[key] == update[key]


def test_post_duplicate_uuid(testapp, award):
    item = {
        'uuid': award['uuid'],
        'name': 'NIS39393',
        'title': 'Grant to make snow',
        'project': 'ENCODE',
        'rfa': 'ENCODE3',
    }
    testapp.post_json('/award', item, status=409)


def test_user_effective_principals(submitter, lab, anontestapp, execute_counter):
    email = submitter['email']
    with execute_counter.expect(1):
        res = anontestapp.get('/@@testing-user',
                              extra_environ={'REMOTE_USER': str(email)})
    assert sorted(res.json['effective_principals']) == [
        'group.submitter',
        'lab.%s' % lab['uuid'],
        'remoteuser.%s' % email,
        'submits_for.%s' % lab['uuid'],
        'system.Authenticated',
        'system.Everyone',
        'userid.%s' % submitter['uuid'],
        'viewing_group.ENCODE',
    ]


def test_page_toplevel(workbook, anontestapp):
    res = anontestapp.get('/test-section/', status=200)
    assert res.json['@id'] == '/test-section/'

    res = anontestapp.get('/pages/test-section/', status=301)
    assert res.location == 'http://localhost/test-section/'


def test_page_nested(workbook, anontestapp):
    res = anontestapp.get('/test-section/subpage/', status=200)
    assert res.json['@id'] == '/test-section/subpage/'


def test_page_nested_in_progress(workbook, anontestapp):
    return anontestapp.get('/test-section/subpage-in-progress/', status=403)


def test_page_homepage(workbook, anontestapp):
    res = anontestapp.get('/pages/homepage/', status=200)
    assert res.json['canonical_uri'] == '/'

    res = anontestapp.get('/', status=200)
    assert 'default_page' in res.json
    assert res.json['default_page']['@id'] == '/pages/homepage/'


def test_page_collection_default(workbook, anontestapp):
    res = anontestapp.get('/pages/images/', status=200)
    assert res.json['canonical_uri'] == '/images/'

    res = anontestapp.get('/images/', status=200)
    assert 'default_page' in res.json
    assert res.json['default_page']['@id'] == '/pages/images/'


def test_jsonld_context(testapp):
    res = testapp.get('/terms/')
    assert res.json


def test_jsonld_term(testapp):
    res = testapp.get('/terms/submitted_by')
    assert res.json


@pytest.mark.slow
@pytest.mark.parametrize('item_type', TYPE_LENGTH)
def test_index_data_workbook(workbook, testapp, indexer_testapp, item_type):
    res = testapp.get('/%s?limit=all' % item_type).follow(status=200)
    for item in res.json['@graph']:
        indexer_testapp.get(item['@id'] + '@@index-data')


@pytest.mark.parametrize('item_type', TYPE_LENGTH)
def test_profiles(testapp, item_type):
    from jsonschema_serialize_fork import Draft4Validator
    res = testapp.get('/profiles/%s.json' % item_type).maybe_follow(status=200)
    errors = Draft4Validator.check_schema(res.json)
    assert not errors


def test_bad_frame(testapp, award):
    res = testapp.get(award['@id'] + '?frame=bad', status=404)
    assert res.json['detail'] == '?frame=bad'



from uuid import UUID
from .schema_utils import validate_request
from .validation import ValidationFailure


# No-validation validators


def no_validate_item_content_post(context, request):
    data = request.json
    request.validated.update(data)


def no_validate_item_content_put(context, request):
    data = request.json
    if 'uuid' in data:
        if UUID(data['uuid']) != context.uuid:
            msg = 'uuid may not be changed'
            raise ValidationFailure('body', ['uuid'], msg)
    request.validated.update(data)


def no_validate_item_content_patch(context, request):
    data = context.properties.copy()
    data.update(request.json)
    if 'uuid' in data:
        if UUID(data['uuid']) != context.uuid:
            msg = 'uuid may not be changed'
            raise ValidationFailure('body', ['uuid'], msg)
    request.validated.update(data)


# Schema checking validators


def validate_item_content_post(context, request):
    data = request.json
    validate_request(context.type_info.schema, request, data)


def validate_item_content_put(context, request):
    data = request.json
    schema = context.type_info.schema
    if 'uuid' in data and UUID(data['uuid']) != context.uuid:
        msg = 'uuid may not be changed'
        raise ValidationFailure('body', ['uuid'], msg)
    accession = context.properties.get('accession')
    if accession and accession != data.get('accession'):
        msg = 'must specify original accession'
        raise ValidationFailure('body', ['accession'], msg)
    current = context.upgrade_properties().copy()
    current['uuid'] = str(context.uuid)
    validate_request(schema, request, data, current)


def validate_item_content_patch(context, request):
    data = context.upgrade_properties().copy()
    if 'schema_version' in data:
        del data['schema_version']
    data.update(request.json)
    schema = context.type_info.schema
    if 'uuid' in data and UUID(data['uuid']) != context.uuid:
        msg = 'uuid may not be changed'
        raise ValidationFailure('body', ['uuid'], msg)
    current = context.upgrade_properties().copy()
    current['uuid'] = str(context.uuid)
    validate_request(schema, request, data, current)



import pytest


def included(config):
    def new_item_search_config():
        return {
            'facets': {'a': 'b'}
        }
    config.register_search_config(
        'OtherConfigItem', new_item_search_config
    )
    

def test_searches_configs_search_config_decorator(config, dummy_request):
    from snovault.elasticsearch.searches.interfaces import SEARCH_CONFIG
    from snovault.elasticsearch.searches.configs import search_config
    assert dummy_request.registry[SEARCH_CONFIG].get('TestConfigItem').facets == {'a': 'b'}
    config.include('snovault.elasticsearch.searches.configs')
    config.include(included)
    config.commit()
    assert config.registry[SEARCH_CONFIG].registry.get('OtherConfigItem').facets == {'a': 'b'}
    config.register_search_config('OtherConfigItem', lambda: {'facets': {'c': 'd'}})
    config.commit()
    assert config.registry[SEARCH_CONFIG].registry.get('OtherConfigItem').facets == {'c': 'd'}



from snovault.util import get_root_request
from elasticsearch.helpers import scan
from pyramid.threadlocal import get_current_request
from zope.interface import alsoProvides
from .interfaces import (
    ELASTIC_SEARCH,
    ICachedItem,
    RESOURCES_INDEX,
)


SEARCH_MAX = (2 ** 31) - 1


def includeme(config):
    from snovault import STORAGE
    registry = config.registry
    es = registry[ELASTIC_SEARCH]
    es_index = RESOURCES_INDEX
    wrapped_storage = registry[STORAGE]
    registry[STORAGE] = PickStorage(ElasticSearchStorage(es, es_index), wrapped_storage)


def force_database_for_request():
    request = get_current_request()
    if request:
        request.datastore = 'database'


class CachedModel(object):
    def __init__(self, hit):
        self.hit = hit
        self.source = hit['_source']

    @property
    def item_type(self):
        return self.source['item_type']

    @property
    def properties(self):
        return self.source['properties']

    @property
    def propsheets(self):
        return self.source['propsheets']

    @property
    def uuid(self):
        return self.source['uuid']

    @property
    def tid(self):
        return self.source['tid']

    def invalidated(self):
        request = get_root_request()
        if request is None:
            return False
        edits = dict.get(request.session, 'edits', None)
        if edits is None:
            return False
        version = self.hit['_version']
        source = self.source
        linked_uuids = set(source['linked_uuids'])
        embedded_uuids = set(source['embedded_uuids'])
        for xid, updated, linked in edits:
            if xid < version:
                continue
            if not embedded_uuids.isdisjoint(updated):
                return True
            if not linked_uuids.isdisjoint(linked):
                return True
        return False

    def used_for(self, item):
        alsoProvides(item, ICachedItem)


class PickStorage(object):
    def __init__(self, read, write):
        self.read = read
        self.write = write

    def storage(self):
        request = get_current_request()
        if request and request.datastore == 'elasticsearch':
            return self.read
        return self.write

    def get_by_uuid(self, uuid):
        storage = self.storage()
        model = storage.get_by_uuid(uuid)
        if storage is self.read:
            model_invalidated = bool(model and model.invalidated())
            if model_invalidated:
                force_database_for_request()
            if model is None or model_invalidated:
                return self.write.get_by_uuid(uuid)
        return model

    def get_by_unique_key(self, unique_key, name, index=None):
        storage = self.storage()
        model = storage.get_by_unique_key(unique_key, name, index=index)
        if storage is self.read:
            model_invalidated = bool(model and model.invalidated())
            if model_invalidated:
                force_database_for_request()
            if model is None or model_invalidated:
                return self.write.get_by_unique_key(unique_key, name, index=index)
        return model

    def get_rev_links(self, model, rel, *item_types):
        storage = self.storage()
        if isinstance(model, CachedModel) and storage is self.write:
            model = storage.get_by_uuid(str(model.uuid))
        return storage.get_rev_links(model, rel, *item_types)

    def __iter__(self, *item_types):
        return self.storage().__iter__(*item_types)

    def __len__(self, *item_types):
        return self.storage().__len__(*item_types)

    def create(self, item_type, uuid):
        return self.write.create(item_type, uuid)

    def update(self, model, properties=None, sheets=None, unique_keys=None, links=None):
        return self.write.update(model, properties, sheets, unique_keys, links)


class ElasticSearchStorage(object):
    writeable = False

    def __init__(self, es, index):
        self.es = es
        self.index = index

    def _one(self, query, index=None):
        if index is None:
            index = self.index
        data = self.es.search(index=index, body=query)
        hits = data['hits']['hits']
        if len(hits) != 1:
            return None
        model = CachedModel(hits[0])
        return model

    def get_by_uuid(self, uuid):
        query = {
            'query': {
                'term': {
                    'uuid': str(uuid)
                }
            },
            'version': True
        }
        result = self.es.search(index=self.index, body=query, _source=True, size=1)
        if result['hits']['total'] == 0:
            return None
        hit = result['hits']['hits'][0]
        return CachedModel(hit)

    def get_by_unique_key(self, unique_key, name, index=None):
        term = 'unique_keys.' + unique_key
        query = {
            'query': {
                'term': {term: name}
            },
            'version': True,
        }
        return self._one(query, index)

    def get_rev_links(self, model, rel, *item_types):
        filter_ = {'term': {'links.' + rel: str(model.uuid)}}
        if item_types:
            filter_ = [
                filter_,
                {'terms': {'item_type': item_types}},
            ]
        query = {
            'stored_fields': [],
            'query': {
                'bool': {
                    'filter': filter_,
                }
            }
        }

        return [
            hit['_id'] for hit in scan(self.es, query=query)
        ]

    def __iter__(self, *item_types):
        query = {
            'stored_fields': [],
            'query': {
                'bool': {
                    'filter': {'terms': {'item_type': item_types}} if item_types else {'match_all': {}}
                }
            }
        }
        for hit in scan(self.es, query=query):
            yield hit['_id']

    def __len__(self, *item_types):
        query = {
            'filter': {'terms': {'item_type': item_types}} if item_types else {'match_all': {}},
        }
        result = self.es.count(index=self.index, body=query)
        return result['count']



import pytest
from pytest_bdd import (
    scenarios,
    then,
    when,
)
from . import browsersteps

pytestmark = [
    pytest.mark.bdd,
    pytest.mark.usefixtures('workbook'),
]

scenarios('generics.feature', strict_gherkin=False)


# https://github.com/pytest-dev/pytest-bdd/issues/124


@when('I visit "/<type_name>/"')
def i_visit_the_collection_for_type_name(browser, base_url, type_name):
    url = '/{}/'.format(type_name)
    browsersteps.when_i_visit_url(browser, base_url, url)


@when('I click the link with text that contains "<link_text>"')
def click_link_with_text_that_contains_link_text(browser, link_text):
    browsersteps.click_link_with_text_that_contains(browser, link_text)


@then('I should see an element with the css selector ".view-item.type-<type_name>"')
def should_see_element_with_css_type_name(browser, type_name):
    css = ".view-item.type-{}".format(type_name)
    browsersteps.should_see_element_with_css(browser, css)
